import math
from math import exp
import regex as re
import os
import pandas as pd
from sklearn.ensemble import GradientBoostingRegressor
import pickle
import numpy as np
from featurization import featurize_data


def load_model_params(score_name):
    try:
        if score_name == 'cfd':
            mm_scores = pickle.load(open(os.path.dirname(os.path.realpath(__file__)) + '/pkl/mismatch_score.pkl', 'rb'))
            pam_scores = pickle.load(open(os.path.dirname(
                os.path.realpath(__file__)) + '/pkl/PAM_scores.pkl', 'rb'))
            return mm_scores, pam_scores
        if score_name == 'azimuth':
            model = pickle.load(open(os.path.dirname(
                os.path.realpath(__file__)) + '/pkl/python3_V3_model_no.pos.pickle', 'rb'))
            return model
        if score_name == 'doench14':
            doench14params = pickle.load(open(os.path.dirname(
                os.path.realpath(__file__)) + '/pkl/doench14params.pkl', 'rb'))
        return doench14params
    except:
        raise Exception(f"pickle file containing {score_name} could not be opened")

def revcom(s):
    basecomp = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A','U':'A'}
    letters = list(s[::-1])
    letters = [basecomp[base] for base in letters]
    return ''.join(letters)

def doench2014(cas9_sites):
        """
        Doench 2014 'on-target score'
        Input is a 30mer: 4bp 5', 20bp guide, 3bp PAM, 3bp 5'
        Doench et al, Nat Biotech 2014, PMID 25184501, http://www.broadinstitute.org/rnai/public/analysis-tools/sgrna-design
        Code from crispor - https://github.com/maximilianh/crisporWebsite
        """
        doench2014params = load_model_params(score_name = 'doench14')
        scores = []
        global gcWeight
        intercept = 0.59763615
        gcHigh = -0.1665878
        gcLow = -0.2026259

        for seq in cas9_sites:
            assert (len(seq) == 30)
            score = intercept
            guideSeq = seq[4:24]
            gcCount = guideSeq.count("G") + guideSeq.count("C")

            if gcCount <= 10:
                gcWeight = gcLow
            if gcCount > 10:
                gcWeight = gcHigh

            score += abs(10 - gcCount) * gcWeight

            for pos, modelSeq, weight in doench2014params:
                subSeq = seq[pos:pos + len(modelSeq)]
                if subSeq == modelSeq:
                    score += weight
            scores.append(int(100 * (1.0 / (1.0 + math.exp(-score)))))

        return score

def azimuth(cas9_sites):
    '''
    Doench/Fusi 2016 Rule -2 on-target / efficiency score now packaged as 'Azimuth'
    This script is copied and modified to suit from https://github.com/MicrosoftResearch/Azimuth
    predicts whether a guide exhibits strong or weak cleavage
    Score range 0-100. A score higher than 55% is recommended
    '''
    #Doench/Fusi 2016 Rule -2 'on-target score'
    # This script is copied and modified to suit from https://github.com/MicrosoftResearch/Azimuth
    model = load_model_params(score_name='azimuth')
    model, learn_options = model

    res = []
    for seq in cas9_sites:
        if "N" in seq:
            res.append(-1)  # can't do Ns
            continue

        pam = seq[25:27]
        if pam != "GG":
            # res.append(-1)
            # continue
            seq = list(seq)
            seq[25] = "G"
            seq[26] = "G"
            seq = "".join(seq)
        res.append(seq)
    seqs = np.array(res)

    learn_options["V"] = 2

    Xdf = pd.DataFrame(columns=['30mer', 'Strand'],
                       data=zip(seqs, np.repeat('NA', seqs.shape[0])))
    gene_position = pd.DataFrame(columns=['Percent Peptide', 'Amino Acid Cut position'],
                                 data=zip(np.ones(seqs.shape[0]) * -1,
                                          np.ones(seqs.shape[0]) * -1))

    feature_sets = featurize_data(data=Xdf, learn_options=learn_options, Y=pd.DataFrame())
    keys = list(feature_sets.keys())
    N = feature_sets[list(keys)[0]].shape[0]
    inputs = np.zeros((N, 0))
    feature_names = []
    dim = {}
    dimsum = 0
    for set in keys:
        inputs_set = feature_sets[set].values
        dim[set] = inputs_set.shape[1]
        dimsum += dim[set]
        inputs = np.hstack((inputs, inputs_set))
        feature_names.extend(feature_sets[set].columns.tolist())
    scores = model.predict(inputs)
    scores = [(s * 100).round(2) for s in scores]
    return scores


def oofscore(seq):
    '''
    copied and adapted code from Bae et al. https://www.nature.com/articles/nmeth.3015
    computes both microhomology and out-of-frame score
    A measurement of how likely an out-of-frame deletion occurs after a knock-out experiment
    based on microhomology
    scoring range 0-100. The higher the oof score, the more deletions have a length that is not a multiple of three
     A score above 66 is recommended
    The higher the oof score, the more deletions have a length that is not a multiple of three
    '''
    length_weight = 20.0
    left = 30
    right = len(seq) - int(left)

    s1 = []
    for k in range(2, left)[::-1]:
        for j in range(left, left + right - k + 1):
            for i in range(0, left - k + 1):
                if seq[i:i + k] == seq[j:j + k]:
                    length = j - i
                    s1.append(seq[i:i + k] + '\t' + str(i) + '\t' + str(i + k) + '\t' + str(j) + '\t' + str(j + k) + '\t' + str(length))

    if s1 != "":
        list_f1 = s1
        sum_score_3 = 0
        sum_score_not_3 = 0

        for i in range(len(list_f1)):
            n = 0
            score_3 = 0
            score_not_3 = 0
            line = list_f1[i].split('\t')
            scrap = line[0]
            left_start = int(line[1])
            left_end = int(line[2])
            right_start = int(line[3])
            right_end = int(line[4])
            length = int(line[5])

            for j in range(i):
                line_ref = list_f1[j].split('\t')
                left_start_ref = int(line_ref[1])
                left_end_ref = int(line_ref[2])
                right_start_ref = int(line_ref[3])
                right_end_ref = int(line_ref[4])

                if (left_start >= left_start_ref) and (left_end <= left_end_ref) and (
                        right_start >= right_start_ref) and (right_end <= right_end_ref):
                    if (left_start - left_start_ref) == (right_start - right_start_ref) and (
                            left_end - left_end_ref) == (right_end - right_end_ref):
                        n += 1
                else:
                    pass

            if n == 0:
                if (length % 3) == 0:
                    length_factor = round(1 / exp((length) / (length_weight)), 3)
                    num_GC = len(re.findall('G', scrap)) + len(re.findall('C', scrap))
                    score_3 = 100 * length_factor * ((len(scrap) - num_GC) + (num_GC * 2))

                elif (length % 3) != 0:
                    length_factor = round(1 / exp((length) / (length_weight)), 3)
                    num_GC = len(re.findall('G', scrap)) + len(re.findall('C', scrap))
                    score_not_3 = 100 * length_factor * ((len(scrap) - num_GC) + (num_GC * 2))
            sum_score_3 += score_3
            sum_score_not_3 += score_not_3

        mh_score = round(sum_score_3 + sum_score_not_3,2)
        oof_score = round((sum_score_not_3) * 100 / (sum_score_3 + sum_score_not_3),2)

    return mh_score, oof_score


def cfd_score(seq1, seq2,pam,mm_scores,pam_scores):
    '''
    Doench 2016 off-target scoring
    Doench, Fusi, et al.  Nature Biotechnology 34, 184–191 (2016)."

    '''
    score = 1
    shorter, longer = sorted([seq1, seq2], key=len)
    for i in range(-len(shorter), 0):
        if (seq1[i] != seq2[i]):
            key = 'r' + seq1[i] + ':d' + revcom(seq2[i]) + ',' + str(20 + i + 1)
            score *= mm_scores[key]
    score *= pam_scores[pam[-2:]]
    return score

def cfd_multi_score(spacer, off_target_seqs):
    '''
    on_target_seq is spacer site and off_target includes pam
    '''
    mm_scores, pam_scores = load_model_params(score_name='cfd')
    scores = []
    seq1 = spacer.upper().replace('T', 'U')
    sum_scores = 0

    for off_target_seq in off_target_seqs:
        m_off = re.search('[^ATCG]',off_target_seq)
        if (m_off is None):
            pam = off_target_seq[-2:]
            seq2 = off_target_seq[:20].upper().replace('T','U')
            if off_target_seq != spacer:
                score = cfd_score(seq1, seq2, pam,mm_scores,pam_scores)
                sum_scores += score
            else:
                score = -1
        scores.append(round(score,2))
        guide_cfd_score = 100 / (100+sum_scores)
        guide_cfd_score = int(round(guide_cfd_score*100))
    return scores, guide_cfd_score



'''
Test
pam = 'AGG'
spacer = 'GTGCGGCTGGCCCAGGACCT'
target = spacer + pam
target30mer = 'CTTGTGCGGCTGGCCCAGGACCTAGGCGAG'
target60mer = 'ATCTCTTACAACGACTTCTTGTGCGGCTGGCCCAGGACCTAGGCGAGGCAGTAGGGGATGACA'
off_target_seqs = ['GTGGGGCTGACCCAGGACCTGAG','GGGCCTCTGGCCCAGGACCTGGG']


print('Microhomology Out-Of-Frame score: ',oofscore(seq = target60mer))
# Ans: 3730.1, 64.2

print('Azimuth on-target score: ',azimuth([target30mer])[0])
#Ans: 0.38

print('CFD scores, Guide specifity score: ',cfd_multi_score(spacer=spacer, off_target_seqs = off_target_seqs))
#Ans: [0.12, 0.44], 99
'''




import pickle
import pandas as pd
from subprocess import Popen
from os import listdir, remove
from collections import defaultdict

### This script takes ~45minutes to run and ~32seconds/guide found
# I recommend no more than 100 guides run at a time
'''
Cas-OFFinder 3.0.0 beta (Jan 24 2021)
Copyright (c) 2021 Jeongbin Park and Sangsu Bae
Website: http://github.com/snugel/cas-offinder

Usage: cas-offinder {input_filename|-} {C|G|A}[device_id(s)] {output_filename|-}
(C: using CPUs, G: using GPUs, A: using accelerators)
'''


def make_casoffinder_input(infile,fasta_fname,pam, pamISfirst, guidelen,guides,gnames,casoff_params):
    ## create input file for cas-offinder
    mm, RNAbb, DNAbb, PU = casoff_params

    with open(infile, 'w') as f:
        f.writelines(fasta_fname + "\n")
        line = 'N' * guidelen

        if pamISfirst:
            line = f"{pam}{line} {DNAbb} {RNAbb}" + "\n"
        else:
            line = f"{line}{pam} {DNAbb} {RNAbb}" + "\n"
        f.writelines(line)
        print(line)

        dpam = 'N' * len(pam)
        for grna, gname in zip(guides, gnames):
            if pamISfirst:
                line = f"{dpam}{grna} {mm} {gname}" + "\n"
            else:
                line = f"{grna}{dpam} {mm} {gname}" + "\n"
            f.writelines(line)
            print(line)


def cas_offinder_bulge(input_filename, output_filename,cas_off_expath,bulge):
    '''
     The cas-offinder off-line package contains a bug that doesn't allow bulges
    This script is partially a wrapper for cas-offinder to fix this bug
     created by...
    https://github.com/hyugel/cas-offinder-bulge

    '''
    fnhead = input_filename.replace("_input.txt", "")
    id_dict = {}
    if bulge == True:
        with open(input_filename) as f:
            chrom_path = f.readline()
            pattern, bulge_dna, bulge_rna = f.readline().strip().split()
            isreversed = False
            for i in range(int(len(pattern) / 2)):
                if pattern[i] == 'N' and pattern[len(pattern) - i - 1] != 'N':
                    isreversed = False
                    break
                elif pattern[i] != 'N' and pattern[len(pattern) - i - 1] == 'N':
                    isreversed = True
                    break
            bulge_dna, bulge_rna = int(bulge_dna), int(bulge_rna)
            targets = [line.strip().split() for line in f]
            rnabulge_dic = defaultdict(lambda: [])
            bg_tgts = defaultdict(lambda: set())
            for raw_target, mismatch, gid in targets:
                if isreversed:
                    target = raw_target.lstrip('N')
                    len_pam = len(raw_target) - len(target)
                    bg_tgts['N' * len_pam + target + 'N' * bulge_dna].add(mismatch)
                    id_dict['N' * len_pam + target + 'N' * bulge_dna] = gid
                    for bulge_size in range(1, bulge_dna+1):
                        for i in range(1, len(target)):
                            bg_tgt = 'N' * len_pam + target[:i] + 'N' * bulge_size + target[i:] + 'N' * (bulge_dna - bulge_size)
                            bg_tgts[bg_tgt].add(mismatch)
                            id_dict[bg_tgt] = gid

                    for bulge_size in range(1, bulge_rna+1):
                        for i in range(1, len(target)-bulge_size):
                            bg_tgt = 'N' * len_pam + target[:i] + target[i+bulge_size:] + 'N' * (bulge_dna + bulge_size)
                            bg_tgts[bg_tgt].add(mismatch)
                            rnabulge_dic[bg_tgt].append((i, int(mismatch), target[i:i+bulge_size]))
                            id_dict[bg_tgt] = gid
                else:
                    target = raw_target.rstrip('N')
                    len_pam = len(raw_target) - len(target)
                    bg_tgts['N' * bulge_dna + target + 'N' * len_pam].add(mismatch)
                    id_dict['N' * bulge_dna + target + 'N' * len_pam] = gid
                    for bulge_size in range(1, bulge_dna+1):
                        for i in range(1, len(target)):
                            bg_tgt = 'N' * (bulge_dna - bulge_size) + target[:i] + 'N' * bulge_size + target[i:] + 'N' * len_pam
                            bg_tgts[bg_tgt].add(mismatch)
                            id_dict[bg_tgt] = gid

                    for bulge_size in range(1, bulge_rna+1):
                        for i in range(1, len(target)-bulge_size):
                            bg_tgt = 'N' * (bulge_dna + bulge_size) + target[:i] + target[i+bulge_size:] + 'N' * len_pam
                            bg_tgts[bg_tgt].add(mismatch)
                            rnabulge_dic[bg_tgt].append( (i, int(mismatch), target[i:i+bulge_size]) )
                            id_dict[bg_tgt] = gid
            if isreversed:
                seq_pam = pattern[:len_pam]
            else:
                seq_pam = pattern[-len_pam:]
        with open(fnhead + '_bulge.txt', 'w') as f:
            f.write(chrom_path)
            if isreversed:
                f.write(pattern + bulge_dna*'N' + '\n')
            else:
                f.write(bulge_dna*'N' + pattern + '\n')
            cnt = 0
            for tgt, mismatch in bg_tgts.items():
                f.write(tgt + ' ' + str(max(mismatch)) + ' ' + '\n')
                cnt+=1
        casin = fnhead + '_bulge.txt'
    else:
        nobulge_dict = {}
        with open(input_filename) as inf:
            for line in inf:
                entry = line.strip().split(' ')
                if len(entry) > 2 and len(entry[-1]) > 3:
                    seq, mm, gid = entry
                    nobulge_dict[seq] = [gid, mm]
        casin = input_filename

    print("Created temporary file (%s)." % (casin))
    outfn = fnhead + '_temp.txt'
    print("Running Cas-OFFinder (output file: %s)..." % outfn)
    p = Popen([cas_off_expath, casin, 'C', outfn])
    ret = p.wait()
    if ret != 0:
        print("Cas-OFFinder process was interrupted!")
        exit(ret)
    print("Processing output file...")

    with open(outfn) as fi, open(output_filename, 'w') as fo:
        fo.write('Guide_ID\tBulge type\tcrRNA\tDNA\tChromosome\tPosition\tDirection\tMismatches\tBulge Size\n')
        for line in fi:
            entries = line.strip().split('\t')
            ncnt = 0

            if bulge == False:
                gid, mm = nobulge_dict[entries[0]]
                fo.write(f'{gid}\tX\t{entries[0]}\t{entries[1]}\t{entries[2]}\t{entries[3]}\t{entries[4]}\t{entries[5]}\t0\n')
            else:
                if isreversed:
                    for c in entries[0][::-1]:
                        if c == 'N':
                            ncnt += 1
                        else:
                            break
                    if ncnt == 0:
                        ncnt = -len(entries[0])
                else:
                    for c in entries[0]:
                        if c == 'N':
                            ncnt += 1
                        else:
                            break

                if entries[0] in rnabulge_dic:
                    gid = id_dict[entries[0]]
                    for pos, query_mismatch, seq in rnabulge_dic[entries[0]]:
                        if isreversed:
                            tgt = (seq_pam + entries[0][len_pam:len_pam+pos] + seq + entries[0][len_pam+pos:-ncnt], entries[3][:len_pam+pos] + '-'*len(seq) + entries[3][len_pam+pos:-ncnt])
                        else:
                            tgt = (entries[0][ncnt:ncnt+pos] + seq + entries[0][ncnt+pos:-len_pam] + seq_pam, entries[3][ncnt:ncnt+pos] + '-'*len(seq) + entries[3][ncnt+pos:])
                        if query_mismatch >= int(entries[5]):
                            fo.write('{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\n'.format(gid,'RNA', tgt[0], tgt[1], entries[1], int(entries[2]) + (ncnt if (not isreversed and entries[4] == "+") or (isreversed and ncnt > 0 and entries[4] == "-") else 0), entries[4], int(entries[5]), len(seq)))
                else:
                    gid = id_dict[entries[0]]
                    nbulge = 0
                    if isreversed:
                        for c in entries[0][:-ncnt][len_pam:]:
                            if c == 'N':
                                nbulge += 1
                            elif nbulge != 0:
                                break
                        tgt = (seq_pam + entries[0][:-ncnt][len_pam:].replace('N', '-'), entries[3][:-ncnt])
                    else:
                        for c in entries[0][ncnt:][:-len_pam]:
                            if c == 'N':
                                nbulge += 1
                            elif nbulge != 0:
                                break
                        tgt = (entries[0][ncnt:][:-len_pam].replace('N', '-') + seq_pam, entries[3][ncnt:])
                    fo.write('{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\n'.format(gid,'X' if nbulge == 0 else 'DNA', tgt[0], tgt[1], entries[1], int(entries[2]) + (ncnt if (not isreversed and entries[4] == "+") or (isreversed and ncnt > 0 and entries[4] == "-") else 0), entries[4], int(entries[5]), nbulge))

        remove(fnhead + '_temp.txt')


def agg_results(output_filename,mmco):
    ots_dict = {}
    #grps =[[f'X_{n}'] for n in range(0,mmco)]
    with open(output_filename)as of:
        for line_out in of:
            entry = line_out.strip().split('\t')
            if entry[0] != 'Guide_ID':
                gid, btype, mm, bsize = entry[0], entry[1], entry[7], entry[8]
                if gid not in ots_dict.keys():
                    ots_dict[gid] ={}
                    for i in ['X','RNA','DNA']:
                        for j in range(mmco+1):
                            ots_dict[gid][(i,j)]= 0
                ots_dict[gid][(btype,int(mm))] += 1
    return ots_dict


def write_out_res(ots,gdf,casoff_params,resultsfolder,guide_tab_fname):
    df = pd.DataFrame(ots)

    #this is just a precautionary if this script was run more than once
    if 'Number of Mismatches' in gdf.columns:
        gdf = gdf.drop(columns='Number of Mismatches')

    #Update main guides table
    tot =df.sum(axis = 0)
    tot =tot.rename('Number of Mismatches')
    tot.index.name = 'Guide_ID'
    gdf = gdf.join(tot, on='Guide_ID')
    gdf.to_csv(guide_tab_fname, index = False)

    ## write out aggregate output
    df = df.reset_index()
    df = df.rename(columns = {'level_1': 'Number of Mismatches', 'level_0': 'BulgeType'})
    if casoff_params[1] == 0:
        df = df.loc[df.BulgeType != 'RNA']
    if casoff_params[2] == 0:
        df = df.loc[df.BulgeType != 'DNA']
    df = df.pivot_table(columns = ['BulgeType','Number of Mismatches'],aggfunc="sum")
    out = resultsfolder + 'Num_Mismatches.txt'
    df.to_csv(out, sep = '\t')


def run_casoffinder(resultsfolder,
                    fasta_fname,
                    guide_tab_fname,
                    search_params,
                    cas_off_expath,
                    genome_name,
                    guides_src_name,
                    casoff_params):
    gdf = pd.read_csv(guide_tab_fname)
    ots = {}
    gpr = gdf.groupby('Editor')
    if casoff_params[1:3] == (0, 0):
        bulge = False
    else:
        bulge = True
    # for each editor type find off_targets
    for editor, stats in gpr:
        infile = f"{resultsfolder}{genome_name}_{guides_src_name}_{editor}_casoffinder_input.txt"
        pam, pamISfirst, guidelen = search_params[editor][0:3]
        guides, gnames = list(stats.gRNA), list(stats.Guide_ID)

        # make input file
        make_casoffinder_input(infile,
                               fasta_fname,
                               pam,
                               pamISfirst,
                               guidelen,
                               guides,
                               gnames,
                               casoff_params)

        #start = time.time()
        ## with defualt setting 3mm, 1 dnabulge, 0 rnabulge,
        #run cas-offinder/ adjust input file for bulge
        output_filename = infile.replace('_input.txt', '_output.txt')
        cas_offinder_bulge(infile, output_filename, cas_off_expath, bulge)
        #end = time.time()
        #print(f'total_time for {editor}: {end - start}')
        #print(f'total_time per guide: {(end - start)/len(guides)}')

        #sum off-targets
        ot_dict = agg_results(output_filename,casoff_params[0])
        for k, v in ot_dict.items():
            ots[k] = v
    write_out_res(ots, gdf, casoff_params, resultsfolder, guide_tab_fname)


def main():
    # SNAKEMAKE IMPORTS
    # === Inputs ===
    guides_report = str(snakemake.input.guides_report_out)
    fasta_ref = str(snakemake.input.fasta_ref)
    guide_search_params = str(snakemake.input.guide_search_params)
    snv_site_info = str(snakemake.input.snv_site_info)
    # === Outputs ===
    casoff_out = str(snakemake.output.casoff_out)
    # === Params ===
    RNAbb = str(snakemake.params.rna_bulge)
    DNAbb = str(snakemake.params.dna_bulge)
    mm = str(snakemake.params.max_mismatch)
    PU = 'C'  # G = GPU C = CPU A = Accelerators -- I
    # === Wildcards ===
    guideref_name = str(snakemake.wildcards.guideref_name)
    fastaref_name = str(snakemake.wildcards.fastaref_name)

    '''
    ### For Daniel to snakemake <---------------

    input paths for this script:
        -resultsfolder -- results output folder
        -guide_search_params -- search paramters used in fetchguides
        -guide_tab_fname -- original guide table output from FetchGuides OR ALT process_genomes files
        -fasta_fname -- Hg38 fasta or if using alternative consensus genome
        -(maybe?) casoffinder path

    input variables for the script:
        -genome_name -- name of fasta/consensus we are searching
        -guides_src_name -- name of the guides source genome ex. HG38 or HG02257

    *possibly allow for changes in the cas-offinder parameters
    see bottom of page
    '''

    resultsfolder = "/groups/clinical/projects/editability/medit_queries/medit_test/test_out/"

    paths = listdir(resultsfolder)

    # Guide search params
    search_params = pickle.load(open(guide_search_params, 'rb'))

    # hg38 or consensus sequence
    fasta_fname = fasta_ref
    genome_name = fastaref_name

    # hg38 guides found (but could be {alt_genome}_differences.csv
    guide_tab_fname = guides_report
    guides_src_name = guideref_name

    ### Daniel---> Pycharm is not find subprocess.Popen(casoffinder...) without an absolute path. so I'm adding this
    # but I don't think its needed in the final version
    cas_off_expath = '/home/thudson/miniconda3/envs/edit/bin/cas-offinder'

    # defaults - we may allow users to change these cas-offinder settings?
    # according to Gorodkin et al. and Lin et al.  DNA bulges are even more tolerated than mismatches alone
    # https://www.nature.com/articles/s41467-022-30515-0
    # RNAbb = 0  # RNA bulge, a deletion in the off-target
    # DNAbb = 1  # DNA bulge, an insertion in the off-target
    # mm = 3  # max allowable mismatch
    # PU = 'C'  # G = GPU C = CPU A = Accelerators -- I don't really know which should be default?
    casoff_params = (mm, RNAbb, DNAbb, PU)

    run_casoffinder(resultsfolder,
                    fasta_fname,
                    guide_tab_fname,
                    search_params,
                    cas_off_expath,
                    genome_name,
                    guides_src_name,
                    casoff_params)


if __name__ == "__main__":
    main()



# This file is copied and modified from https://github.com/MicrosoftResearch/Azimuth
# This code was modified to use with python 3 and needed for percent and position scoring
# was removed to simplify code
# Azimuth license applies to this file

from math import exp
from re import findall
import pandas as pd
import Bio.Seq as Seq
import numpy as np
import itertools
import Bio.SeqUtils.MeltingTemp as Tm

def countGC(s, length_audit=True):
    '''
    GC content for only the 20mer, as per the Doench paper/code
    '''
    if length_audit:
        assert len(s) == 30, "seems to assume 30mer"
    return len(s[4:24].replace('A', '').replace('T', ''))


def nucleotide_features(s, order, include_pos_independent, max_index_to_use, prefix="", feature_type='all',
                        raw_alphabet=['A', 'T', 'C', 'G']):
    '''
    compute position-specific order-mer features for the 4-letter alphabet
    (e.g. for a sequence of length 30, there are 30*4 single nucleotide features
          and (30-1)*4^2=464 double nucleotide features
    '''
    assert feature_type in ['all', 'pos_independent', 'pos_dependent']
    if max_index_to_use <= len(s):
        # print "WARNING: trimming max_index_to use down to length of string=%s" % len(s)
        max_index_to_use = len(s)
    if max_index_to_use is not None:
        s = s[:max_index_to_use]
    # assert(len(s)==30, "length not 30")
    # s = s[:30] #cut-off at thirty to clean up extra data that they accidentally left in, and were instructed to ignore in this way
    alphabet = get_alphabet(order, raw_alphabet=raw_alphabet)
    features_pos_dependent = np.zeros(len(alphabet) * (len(s) - (order - 1)))
    features_pos_independent = np.zeros(np.power(len(raw_alphabet), order))

    index_dependent = []
    index_independent = []

    for position in range(0, len(s) - order + 1, 1):
        for l in alphabet:
            index_dependent.append('%s%s_%d' % (prefix, l, position))

    for l in alphabet:
        index_independent.append('%s%s' % (prefix, l))

    for position in range(0, len(s) - order + 1, 1):
        nucl = s[position:position + order]
        features_pos_dependent[alphabet.index(nucl) + (position * len(alphabet))] = 1.0
        features_pos_independent[alphabet.index(nucl)] += 1.0

        # this is to check that the labels in the pd df actually match the nucl and position
        assert index_dependent[alphabet.index(nucl) + (position * len(alphabet))] == '%s%s_%d' % (
        prefix, nucl, position)
        assert index_independent[alphabet.index(nucl)] == '%s%s' % (prefix, nucl)

    # index_independent = ['%s_pi.Order%d_P%d' % (prefix, order,i) for i in range(len(features_pos_independent))]
    # index_dependent = ['%s_pd.Order%d_P%d' % (prefix, order, i) for i in range(len(features_pos_dependent))]

    if np.any(np.isnan(features_pos_dependent)):
        raise Exception("found nan features in features_pos_dependent")
    if np.any(np.isnan(features_pos_independent)):
        raise Exception("found nan features in features_pos_independent")

    if feature_type == 'all' or feature_type == 'pos_independent':
        if feature_type == 'all':
            res = pd.Series(features_pos_dependent, index=index_dependent), pd.Series(features_pos_independent,
                                                                                      index=index_independent)
            assert not np.any(np.isnan(res.values))
            return res
        else:
            res = pd.Series(features_pos_independent, index=index_independent)
            assert not np.any(np.isnan(res.values))
            return res

    res = pd.Series(features_pos_dependent, index=index_dependent)
    assert not np.any(np.isnan(res.values))
    return res


def Tm_feature(data, pam_audit=True, learn_options=None):
    '''
    assuming '30-mer'is a key
    get melting temperature features from:
        0-the 30-mer ("global Tm")
        1-the Tm (melting temperature) of the DNA:RNA hybrid from positions 16 - 20 of the sgRNA, i.e. the 5nts immediately proximal of the NGG PAM
        2-the Tm of the DNA:RNA hybrid from position 8 - 15 (i.e. 8 nt)
        3-the Tm of the DNA:RNA hybrid from position 3 - 7  (i.e. 5 nt)
    '''
    if learn_options is None or 'Tm segments' not in learn_options.keys():
        segments = [(19, 24), (11, 19), (6, 11)]
    else:
        segments = learn_options['Tm segments']
    sequence = data['30mer'].values
    featarray = np.ones((sequence.shape[0], 4))
    for i, seq in enumerate(sequence):
        if seq[25:27] != "GG":
            raise Exception("expected GG but found %s" % seq[25:27])
        rna = False
        featarray[i, 0] = Tm.Tm_NN(seq)  # 30mer Tm
        featarray[i, 1] = Tm.Tm_NN(seq[segments[0][0]:segments[0][1]], )  # 5nts immediately proximal of the NGG PAM
        featarray[i, 2] = Tm.Tm_NN(seq[segments[1][0]:segments[1][1]])  # 8-mer
        featarray[i, 3] = Tm.Tm_NN(seq[segments[2][0]:segments[2][1]])  # 5-mer

    feat = pd.DataFrame(featarray, index=data.index,
                        columns=["Tm global_%s" % rna, "5mer_end_%s" % rna, "8mer_middle_%s" % rna,
                                 "5mer_start_%s" % rna])

    return feat


def get_alphabet(order, raw_alphabet=['A', 'T', 'C', 'G']):
    alphabet = ["".join(i) for i in itertools.product(raw_alphabet, repeat=order)]
    return alphabet


def NGGX_interaction_feature(data, pam_audit=True):
    '''
    assuming 30-mer, grab the NGGX _ _ positions, and make a one-hot
    encoding of the NX nucleotides yielding 4x4=16 features
    '''
    sequence = data['30mer'].values
    feat_NX = pd.DataFrame()
    # check that GG is where we think
    for seq in sequence:
        if seq[25:27] != "GG":
            raise Exception("expected GG but found %s" % seq[25:27])
        NX = seq[24] + seq[27]
        NX_onehot = nucleotide_features(NX, order=2, feature_type='pos_dependent', max_index_to_use=2, prefix="NGGX")
        # NX_onehot[:] = np.random.rand(NX_onehot.shape[0]) ##TESTING RANDOM FEATURE
        feat_NX = pd.concat([feat_NX, NX_onehot], axis=1)
    return feat_NX.T


def apply_nucleotide_features(seq_data_frame, order, num_proc, include_pos_independent, max_index_to_use, prefix=""):
    if include_pos_independent:
        feat_pd = seq_data_frame.apply(nucleotide_features, args=(order, max_index_to_use, prefix, 'pos_dependent'))
        feat_pi = seq_data_frame.apply(nucleotide_features, args=(order, max_index_to_use, prefix, 'pos_independent'))
        assert not np.any(np.isnan(feat_pd)), "nans here can arise from sequences of different lengths"
        assert not np.any(np.isnan(feat_pi)), "nans here can arise from sequences of different lengths"
        return feat_pd, feat_pi
    else:
        feat_pd = seq_data_frame.apply(nucleotide_features, args=(order, max_index_to_use, prefix, 'pos_dependent'))
        assert not np.any(np.isnan(feat_pd)), "found nan in feat_pd"
        return feat_pd


def nucleotide_features(s, order, max_index_to_use, prefix="", feature_type='all', raw_alphabet=['A', 'T', 'C', 'G']):
    '''
    compute position-specific order-mer features for the 4-letter alphabet
    (e.g. for a sequence of length 30, there are 30*4 single nucleotide features
          and (30-1)*4^2=464 double nucleotide features
    '''
    assert feature_type in ['all', 'pos_independent', 'pos_dependent']
    if max_index_to_use <= len(s):
        # print "WARNING: trimming max_index_to use down to length of string=%s" % len(s)
        max_index_to_use = len(s)

    if max_index_to_use is not None:
        s = s[:max_index_to_use]
    # assert(len(s)==30, "length not 30")
    # s = s[:30] #cut-off at thirty to clean up extra data that they accidentally left in, and were instructed to ignore in this way
    alphabet = get_alphabet(order, raw_alphabet=raw_alphabet)
    features_pos_dependent = np.zeros(len(alphabet) * (len(s) - (order - 1)))
    features_pos_independent = np.zeros(np.power(len(raw_alphabet), order))

    index_dependent = []
    index_independent = []

    for position in range(0, len(s) - order + 1, 1):
        for l in alphabet:
            index_dependent.append('%s%s_%d' % (prefix, l, position))

    for l in alphabet:
        index_independent.append('%s%s' % (prefix, l))

    for position in range(0, len(s) - order + 1, 1):
        nucl = s[position:position + order]
        features_pos_dependent[alphabet.index(nucl) + (position * len(alphabet))] = 1.0
        features_pos_independent[alphabet.index(nucl)] += 1.0

        # this is to check that the labels in the pd df actually match the nucl and position
        assert index_dependent[alphabet.index(nucl) + (position * len(alphabet))] == '%s%s_%d' % (
        prefix, nucl, position)
        assert index_independent[alphabet.index(nucl)] == '%s%s' % (prefix, nucl)

    if np.any(np.isnan(features_pos_dependent)):
        raise Exception("found nan features in features_pos_dependent")
    if np.any(np.isnan(features_pos_independent)):
        raise Exception("found nan features in features_pos_independent")

    if feature_type == 'all' or feature_type == 'pos_independent':
        if feature_type == 'all':
            res = pd.Series(features_pos_dependent, index=index_dependent), pd.Series(features_pos_independent,
                                                                                      index=index_independent)
            assert not np.any(np.isnan(res.values))
            return res
        else:
            res = pd.Series(features_pos_independent, index=index_independent)
            assert not np.any(np.isnan(res.values))
            return res

    res = pd.Series(features_pos_dependent, index=index_dependent)
    assert not np.any(np.isnan(res.values))
    return res

def check_feature_set(feature_sets):
    '''
    Ensure the # of people is the same in each feature set
    '''
    assert feature_sets != {}, "no feature sets present"

    N = None
    for ft in feature_sets.keys():
        N2 = feature_sets[ft].shape[0]
        if N is None:
            N = N2
        else:
            assert N >= 1, "should be at least one individual"
            assert N == N2, "# of individuals do not match up across feature sets"

    for set in feature_sets.keys():
        if np.any(np.isnan(feature_sets[set])):
            raise Exception("found Nan in set %s" % set)


def get_all_order_nuc_features(data, feature_sets, learn_options, maxorder, max_index_to_use, prefix="", quiet=False):
    for order in range(1, maxorder + 1):
        nuc_features_pd, nuc_features_pi = apply_nucleotide_features(data, order, learn_options["num_proc"],
                                                                     include_pos_independent=True,
                                                                     max_index_to_use=max_index_to_use, prefix=prefix)
        feature_sets['%s_nuc_pd_Order%i' % (prefix, order)] = nuc_features_pd
        if learn_options['include_pi_nuc_feat']:
            feature_sets['%s_nuc_pi_Order%i' % (prefix, order)] = nuc_features_pi
        check_feature_set(feature_sets)



def gc_features(data, audit=True):
    gc_count = data['30mer'].apply(lambda seq: countGC(seq, audit))
    gc_count.name = 'GC count'
    gc_above_10 = (gc_count > 10) * 1
    gc_above_10.name = 'GC > 10'
    gc_below_10 = (gc_count < 10) * 1
    gc_below_10.name = 'GC < 10'
    return gc_above_10, gc_below_10, gc_count


def normalize_features(data, axis):
    '''
    input: Pandas.DataFrame of dtype=np.float64 array, of dimensions
    mean-center, and unit variance each feature
    '''
    data -= data.mean(axis)
    data /= data.std(axis)
    # remove rows with NaNs
    data = data.dropna(1)
    if np.any(np.isnan(data.values)): raise Exception("found NaN in normalized features")
    return data


def normalize_feature_sets(feature_sets):
    '''
    zero-mean, unit-variance each feature within each set
    '''

    print("Normalizing features...")
    new_feature_sets = {}
    for set in feature_sets:
        new_feature_sets[set] = normalize_features(feature_sets[set], axis=0)
        if np.any(np.isnan(new_feature_sets[set].values)):
            raise Exception("found Nan feature values in set=%s" % set)
        assert new_feature_sets[set].shape[1] > 0, "0 columns of features"
    return new_feature_sets


def featurize_data(data, learn_options, Y, pam_audit=True, length_audit=True):
    '''
    data = Xdf
    assumes that data contains the 30mer
    returns set of features from which one can make a kernel for each one
    '''
    all_lens = data['30mer'].apply(len).values
    unique_lengths = np.unique(all_lens)
    num_lengths = len(unique_lengths)
    assert num_lengths == 1, "should only have sequences of a single length, but found %s: %s" % (
    num_lengths, str(unique_lengths))

    feature_sets = {}

    if learn_options["nuc_features"]:
        # spectrum kernels (position-independent) and weighted degree kernels (position-dependent)
        get_all_order_nuc_features(data['30mer'], feature_sets, learn_options, learn_options["order"],
                                   max_index_to_use=30, quiet=False)

    check_feature_set(feature_sets)

    if learn_options["gc_features"]:
        gc_above_10, gc_below_10, gc_count = gc_features(data, length_audit)
        feature_sets['gc_above_10'] = pd.DataFrame(gc_above_10)
        feature_sets['gc_below_10'] = pd.DataFrame(gc_below_10)
        feature_sets['gc_count'] = pd.DataFrame(gc_count)

    if learn_options["include_NGGX_interaction"]:
        feature_sets["NGGX"] = NGGX_interaction_feature(data)

    if learn_options["include_Tm"]:
        feature_sets["Tm"] = Tm_feature(data, pam_audit, learn_options=None)


    check_feature_set(feature_sets)

    reordered_feature_sets = {}
    keys = ['gc_count', '_nuc_pd_Order2', '_nuc_pd_Order1', 'gc_above_10', '_nuc_pi_Order1', '_nuc_pi_Order2', 'Tm',
            'gc_below_10', 'NGGX']
    for k in keys:
        reordered_feature_sets[k] = feature_sets[k]

    return reordered_feature_sets






# Native modules
import re
import os
# Installed Modules
import boto3


def list_objects_in_bucket(bucket_name, version, prefix='', continuation_token=None):
    objects_list = []

    # Create an S3 client
    s3 = boto3.client('s3')

    # Configure the request parameters, including the ContinuationToken
    params = {'Bucket': bucket_name, 'Prefix': prefix}
    if continuation_token:
        params['ContinuationToken'] = continuation_token

    # List objects in the bucket with the specified prefix
    response = s3.list_objects_v2(**params)

    # Iterate through the objects
    for obj in response.get('Contents', []):
        # Only compressed FASTA files are listed
        if re.search(r"{}.fa.gz$".format(version), obj['Key']):
            objects_list.append(str(obj['Key']))

    # Check if there are common prefixes (subdirectories)
    if 'CommonPrefixes' in response:
        for subdir in response['CommonPrefixes']:
            subdir_prefix = subdir['Prefix']
            # Recursively list objects in the subdirectory, passing the ContinuationToken
            subdir_objects, _ = list_objects_in_bucket(bucket_name, version, prefix=subdir_prefix, continuation_token=response.get('NextContinuationToken'))
            objects_list += subdir_objects

    return objects_list, response


def download_objects_from_bucket(bucket_name, objects_list, download_path):
	# Create an S3 client
    s3 = boto3.client('s3')
	for obj_key in objects_list:
		# Specify the local file path for downloading
		local_file_path = os.path.join(download_path, os.path.basename(obj_key))

		# Download the object from S3 to the local path
		s3.download_file(bucket_name, obj_key, local_file_path)
		# Optionally, you can print a message to indicate successful download
		print(f"Downloaded: {obj_key} to {local_file_path}")


def main():
	s3_bucket_name = 'human-pangenomics'
	s3_filepath = 'working/'
	assembly_version = "f1_assembly_v2_genbank"
	local_download_path = '/groups/clinical/projects/clinical_shared_data/hprc/aws/'
	# Replace 'your-bucket-name' with your actual S3 bucket name
	file_list, r = list_objects_in_bucket(s3_bucket_name, assembly_version, s3_filepath)

	# Download the objects to the specified path
	download_objects_from_bucket(s3_bucket_name, file_list, local_download_path)


if __name__ == "__main__":
	main()



# == Native Modules ==
from argparse import ArgumentParser as argp
from argparse import RawTextHelpFormatter
import textwrap


def parse_arguments():
	# -> === Launch argparse parser === <-
	parser = argp(
		prog='mEdit',
		description='',
		# epilog="mEdit is pretty cool, huh? :)",
		usage='%(prog)s [options]',
		formatter_class=RawTextHelpFormatter
	)

	programs = parser.add_subparsers(
		title="== mEdit Programs ==",
		description=textwrap.dedent('''
		mEdit can be operated through a list of different programs.'''),
		dest="program",
	)
	# === Db Setup ===
	dbset_parser = programs.add_parser(
		'db_set',
		help=textwrap.dedent('''
			Setup the necessary background data to run mEdit'''),
		formatter_class=RawTextHelpFormatter
	)
	ref_db_parse = dbset_parser.add_argument_group("== Reference Database Pre-Processing ==")
	ref_db_parse.add_argument('-d',
	                          dest='db_path',
	                          default='.',
	                          help=textwrap.dedent('''
	                          Provide the path where the "mEdit_database" 
	                          directory will be created ahead of the analysis.
	                          Requires ~6.5GB in-disk storage 
	                          [default: ./mEdit_database]'''))
	ref_db_parse.add_argument('-p',
	                          dest='threads',
	                          default='1',
	                          help=textwrap.dedent('''
	                          Provide the number of cores for parallel decompression
	                          of mEdit databases.
	                          '''))
	# === Guide Prediction Program ===
	fguides_parser = programs.add_parser(
		'guide_prediction',
		help=textwrap.dedent('''
			The core mEdit program finds potential guides for
			variants specified on the input by searching a diverse set of
			editors.'''),
		formatter_class=RawTextHelpFormatter
	)
	in_out = fguides_parser.add_argument_group("== Input/Output Options ==")
	in_out.add_argument(
		'-i',
		dest='query_input',
		required=True,
		help=textwrap.dedent('''
			Path to plain text file containing the query (or set of queries) 
			of variant(s) for mEdit analysis. Must be a single nucleotide 
			variation. See --qtype for formatting options.
			''')
	)
	in_out.add_argument(
		'-o',
		dest='output',
		default='medit_analysis',
		help=textwrap.dedent('''
			Path to root directory where mEdit outputs will be stored 
			[default: mEdit_analysis_<jobtag>/]''')
	)
	in_out.add_argument('-d',
	                    dest='db_path',
	                    default='.',
	                    help=textwrap.dedent('''
	                    Provide the path where the "mEdit_database" 
	                    directory was created ahead of the analysis 
	                    using the "db_set" program. 
	                    [default: ./mEdit_database]''')
	                    )
	in_out.add_argument('-j',
	                    dest='jobtag',
	                    help=textwrap.dedent('''
	                    Provide the tag associated with the current mEdit job.
	                    mEdit will generate a random jobtag by default''')
	                    )
	run_params = fguides_parser.add_argument_group("== mEdit Core Parameters ==")
	run_params.add_argument(
		'-m',
		dest='mode',
		default='standard',
		choices=['standard', 'private'],
		help=textwrap.dedent('''
			The MODE option determines how mEdit will run your job. 
			[default = "standard"]
			[1-] "standard": will find and process guides based on a 
			reference human genome assembly along with a diverse set of 
			pangenomes from HPRC.
			[2-] "private": requires a private VCF file and use it to 
			process and find guides.''')
	)
	run_params.add_argument(
		'-g',
		dest='private_genome',
		default=None,
		help=textwrap.dedent('''
			Provide a gunzip compressed VCF file to run mEdit’s 
			private mode''')
	)
	run_params.add_argument(
		'--qtype',
		dest='qtype_request',
		default='hgvs',
		choices=['hgvs', 'coord'],
		help=textwrap.dedent('''
			Set the query type provided to mEdit. [default = "hgvs"]
			[1-] "hgvs": must at least contain the Refseq identifier 
			followed by “:” and the commonly used HGVS nomenclature. 
			Example: NM_000518.5:c.114G>A
			[2-] "coord": must contain hg38 coordinates followed by 
			(ALT>REF). Alleles must be the plus strand.
			Example: chr11:5226778C>T\n''')
	)
	run_params.add_argument(
		'--editor',
		dest='editor_request',
		default='clinical',
		choices=['clinical', 'custom', 'user defined list'],
		help=textwrap.dedent('''
			Pick which set of editors will be used in the mEdit run. 
			[default = "clinical"]
			Use medit.py list to access the arrays of editors currently 
			supported in each category.
			[1-] "clinical": a short list of clinically relevant editors 
			that are either in pre-clinical or clinical trials.
			[2-] "custom": select guide search parameters. This requires a
			 separate input of parameters : ‘pam’, ‘pamISfirst’,’guidelen’
			[3-] "user defined list": - Comma-separated list of editors''')
	)
	run_params.add_argument(
		'--be',
		dest='be_request',
		choices=['off', 'default', 'custom', 'user defined list'],
		default='default',
		help=textwrap.dedent('''
			Add this flag to make mEdit process base-editors. 
			[default = off]''')
	)
	run_params.add_argument(
		'--dry',
		dest='dry_run',
		action='store_true',
		help=textwrap.dedent('''
			Perform a dry run of mEdit.'''))
	cluster_opt = fguides_parser.add_argument_group("== SLURM Options ==")
	cluster_opt.add_argument(
		'-p',
		dest='parallel_processes',
		help=textwrap.dedent('''
				Most processes in mEdit can be submitted to SLURM.
				When submitting mEdit jobs to SLURM, the user can specify
				the number of parallel processes that will be sent to the 
				server [default = 1]''')
	)
	cluster_opt.add_argument(
		'--ncores',
		dest='ncores',
		default=2,
		help=textwrap.dedent('''
			Specify the number of cores through which each parallel process 
			will be computed. [default = 2]''')
	)

	cluster_opt.add_argument(
		'--maxtime',
		dest='maxtime',
		default='1:00:00',
		help=textwrap.dedent('''
			Specify the maximum amount of time allowed for each parallel job.
			Format example: 2 hours -> "2:00:00" [default = 1 hour]''')
	)

	# === Off Target Effect Program ===
	casoff_parser = programs.add_parser(
		'offtargets',
		help=textwrap.dedent('''
			Predict off-target effect for the guides found'''),
		formatter_class=RawTextHelpFormatter
	)
	offtarget_params = casoff_parser.add_argument_group("== Off-Target Parameters ==")
	offtarget_params.add_argument(
		'-r', help='Reference Genome')

	# Parse arguments from the command line
	arguments = parser.parse_args()
	return arguments



# == Native Modules ==
from os.path import abspath
# == Installed Modules ==
import yaml
# == Project Modules ==
from prog.medit_lib import (launch_shell_cmd,
                            list_files_by_extension,
                            pickle_chromosomes,
                            set_export,
                            write_yaml_to_file)


def dbset(args):
	# === Load template configuration file ===
	with open("config/medit_database.yaml", 'r') as config_handle:
		config_db_template = yaml.safe_load(config_handle)

	# === Load Database Path ===
	db_path_full = f"{abspath(args.db_path)}/medit_database"
	config_db_dir_path = f"{db_path_full}/config_db"

	threads = args.threads

	vcf_dir_path = f"{db_path_full}/standard/source_vcfs"
	config_db_path = f"{config_db_dir_path}/config_db.yaml"

	set_export(vcf_dir_path)
	set_export(config_db_dir_path)
	# === Assign Variables to Configuration File ===
	#   == Parent Database Path
	config_db_template['meditdb_path'] = f"{db_path_full}"
	#   == Assign jobtag and Fasta root path ==
	fasta_root_path = f"{db_path_full}/{config_db_template['fasta_root_path']}"
	config_db_template['fasta_root_path'] = fasta_root_path
	#   == Parse the Processed Tables folder and its contents ==
	processed_tables = f"{db_path_full}/{config_db_template['processed_tables']}"
	config_db_template["processed_tables"] = f"{processed_tables}"
	config_db_template["simple_tables"] = f"{processed_tables}/{config_db_template['simple_tables']}"
	config_db_template["hgvs_lookup"] = f"{processed_tables}/{config_db_template['hgvs_lookup']}"
	config_db_template["clinvar_update"] = f"{processed_tables}/{config_db_template['clinvar_update']}"
	config_db_template["refseq_table"] = f"{processed_tables}/{config_db_template['refseq_table']}"

	#   == Parse the Raw Tables folder and its contents ==
	raw_tables = f"{db_path_full}/{config_db_template['raw_tables']}"
	config_db_template["raw_tables"] = f"{raw_tables}"
	config_db_template["clinvar_summary"] = f"{raw_tables}/{config_db_template['clinvar_summary']}"
	config_db_template["hpa"] = f"{raw_tables}/{config_db_template['hpa']}"
	config_db_template["gencode"] = f"{raw_tables}/{config_db_template['gencode']}"

	# === Write YAML configs to mEdit Root Directory ===
	write_yaml_to_file(config_db_template, config_db_path)

	# === Download Data ===
	#   == SeqRecord Pickles
	print("Downloading Database of Genomic References")
	launch_shell_cmd(f"aws s3 cp --recursive s3://medit.db/genome_pkl {fasta_root_path}")
	print("Processing FASTA reference assembly")
	#   == Only one file is expected in this directory. Hence, the 1st item of the list
	reference_fasta_path = list_files_by_extension(fasta_root_path, 'fa.gz')[0]
	launch_shell_cmd(f"bgzip -c -@ {threads} -d {reference_fasta_path} > {fasta_root_path}/tmp_hg38.fa")
	pickle_chromosomes(f"{fasta_root_path}/tmp_hg38.fa", fasta_root_path)
	#   == HPRC VCF files Setup
	launch_shell_cmd(f"aws s3 cp --recursive s3://medit.db/hprc/ {vcf_dir_path}")
	#   == Processed Tables and Raw Tables
	print("Downloading Pre-Processed Background Data Sets")
	launch_shell_cmd(f"aws s3 cp s3://medit.db/processed_tables.tar.gz {db_path_full}")
	launch_shell_cmd(f"aws s3 cp s3://medit.db/raw_tables.tar.gz {db_path_full}")
	print("Decompressing Databases")
	launch_shell_cmd(f"pigz -p {threads} -d {db_path_full}/*.gz")
	launch_shell_cmd(f"tar -xf {db_path_full}/raw_tables.tar --directory={db_path_full}/ && "
	                 f"rm {db_path_full}/raw_tables.tar")
	launch_shell_cmd(f"tar -xf {db_path_full}/processed_tables.tar --directory={db_path_full}/ && "
	                 f"rm {db_path_full}/processed_tables.tar")



# == Native Modules ==
# == Installed Modules ==
# == Project Modules ==
from prog.guide_prediction import guide_prediction as guide_prediction
from prog.db_set import dbset as db_set
from prog.arguments import parse_arguments
from prog.medit_lib import date_tag


def main():
	# === Call argument parsing function ===
	args = parse_arguments()
	# mEdit Program
	program = args.program
	# Assign jobtag and run mode to config
	try:
		jobtag = args.jobtag
	except AttributeError:
		jobtag = date_tag()
		args.user_jobtag = False
	args.user_jobtag = True
	# Run mEdit Programs
	if program == "guide_prediction":
		guide_prediction(args, jobtag)

	# == Database Parameters
	if program == "db_set":
		db_set(args)


if __name__ == "__main__":
	main()



# Native modules
import re
from argparse import ArgumentParser as argp
# Installed modules
import pandas as pd
from Bio import Entrez
import yaml
import urllib.request
import urllib.error
import urllib3
import xmltodict


def parse_arguments():
	#  Launch argparse parser
	parser = argp(
		prog='omim2hgvs',
		description='',
		usage='%(prog)s [options]')
	# Define arguments
	parser.add_argument('omim_list',
	                    help='Path to a plain txt containing a column of omim entries')  # positional argument
	parser.add_argument('-o', default='hgvs_list.txt',
	                    dest='outfile',
	                    help='Path to output the list of HGVSs')
	parser.add_argument('-d',
	                    dest='db_from',
	                    default='clinvar',
	                    choices=['clinvar', 'gene','snp'],
	                    help='The NCBI database where the IDs in the input list will be matched against')
	parser.add_argument('-l',
	                    dest='login_email',
	                    default='thedoudnalab@gmail.com',
	                    help='Email used to connect with NCBI Entrez')

	# Parse arguments from the command line
	arguments = parser.parse_args()
	return arguments


def elink_routine(db, hit_uid):
	dup_check = []
	not_found = ""
	linked = ""
	link_record = ""
	handle = None
	server_attempts = 0
	try:
		handle = Entrez.elink(dbfrom=f"{db}", db='clinvar', id=f"{hit_uid}")
	except urllib.error.HTTPError as err:
		if err.code == 500:
			print(f'An internal server error occurred while handling the accession {hit_uid}')
			not_found = hit_uid
			return linked, hit_uid, not_found
	try:
		link_record = Entrez.read(handle)
	except RuntimeError:
		not_found = hit_uid
	if link_record:
		try:
			linked = []
			# linked = link_record[0]['LinkSetDb'][0]['Link'][0]['Id']
			for link_idx in range(len(link_record[0]['LinkSetDb'][0]['Link'])):
				link_id = link_record[0]['LinkSetDb'][0]['Link'][link_idx]['Id']
				linked.append(link_id)
				if link_id not in dup_check:
					dup_check.append(link_id)
		except (IndexError, KeyError):
			not_found = hit_uid
	handle.close()
	return linked, hit_uid, not_found, link_record


def cross_db_search(query_list, db_name):
	progress = 0
	source2target = {}
	not_found_list = []
	full_record = ''
	for hit in query_list:
		progress += 1
		dup_check = []
		# Standardize identifiers to NCBI UIDs through ESearch
		handle = Entrez.esearch(db=f"{db_name}", term=f"{hit}", idtype="acc")
		search_record = Entrez.read(handle)
		try:
			uid = search_record['IdList'][0]
		except IndexError:
			continue
		handle.close()

		# Loop through databases (found in config) and grab Nuccore UIDs
		if uid in set(dup_check):
			continue
		link_list, loop_nuc_acc, not_found_hit, full_record = elink_routine(db_name, uid)
		if not_found_hit:
			not_found_list.append(not_found_hit)
			continue
		if link_list:
			dup_check.append(uid)
			source2target.setdefault(loop_nuc_acc, link_list)

	return source2target, list(set(not_found_list)), full_record


def url2xml_dict(url):
	http = urllib3.PoolManager()
	file = http.request('GET', str(url))
	data = file.data
	data = xmltodict.parse(data)
	return data


def fetch_hgvs_list(query_list, db_name):
	# Get Genbank records for each Nuccore UID
	hgvs_records = {}
	not_found = []
	for uid in query_list:
		handle = Entrez.esummary(db=db_name, id=f"{uid}", rettype="clinvarset", retmode="default")
		xml_data = url2xml_dict(handle.url)
		try:
			description = xml_data['eSummaryResult']['DocumentSummarySet']['DocumentSummary']['clinical_significance']['description']
			if re.search('pathogenic', description, re.IGNORECASE):
				hgvs = xml_data['eSummaryResult']['DocumentSummarySet']['DocumentSummary']['title']
				# Returns a list  of Genbank SeqRecords objects
				hgvs_records.setdefault(uid, hgvs)
		except KeyError:
			not_found.append(uid)
			continue
		# Returns a list  of Genbank SeqRecords objects
	return hgvs_records, not_found


def main():
	# DEBUG
	# queries = ["264300", "610006", "616034", "246450", "605911"]
	# in_path = "/groups/doudna/projects/daniel_projects/editability/metadata/omim_test.txt"

	args = parse_arguments()
	in_path = args.omim_list
	db_from = args.db_from
	entrez_login = args.login_email
	queries_df = pd.read_csv(in_path, low_memory=False, header=None)
	queries = queries_df.iloc[:, 0].tolist()
	#
	# # Load config file
	# with open("/groups/doudna/projects/daniel_projects/editability/config/id_convert.yaml", "r") as f:
	# 	config = yaml.load(f, Loader=yaml.FullLoader)

	# Entrez authentication
	Entrez.email = entrez_login

	# Query NCBI to get nuccore UIDs associated with the protein hits using ESearch/ELink
	print("Linking protein hit ids to Nuccore entries")
	hit_to_link, hits_not_found, record = cross_db_search(queries, db_from)

	# for key in hit_to_links
	hgvs_list = []
	for input_id in hit_to_link:
		hgvs_list.extend(fetch_hgvs_list(hit_to_link[input_id], 'clinvar'))


if __name__ == "__main__":
	main()



################
#   Keeps Clinvar data up-to-date
#   Checks User input and Terms
###############

import pandas as pd
from datetime import datetime, date
import subprocess
import gzip
import regex as re


class Validator:
    '''
    Validates inputs and datbases
    '''

    clinvar_ftp = "https://ftp.ncbi.nlm.nih.gov/pub/clinvar/tab_delimited/variant_summary.txt.gz"
    clinvar_vcf = ' https://ftp.ncbi.nlm.nih.gov/pub/clinvar/vcf_GRCh38/clinvar.vcf.gz'
    clinvar_index = ' https://ftp.ncbi.nlm.nih.gov/pub/clinvar/vcf_GRCh38/clinvar.vcf.gz.tbi'
    refseq = "https://hgdownload.soe.ucsc.edu/goldenPath/hg38/database/ncbiRefSeq.txt.gz"

    def __init__(self, datadir):

        # database paths
        self.raw_tables = f"{datadir}raw_tables/"
        self.clinvar_summary = f"{self.raw_tables}clinvar/variant_summary.txt.gz"  # raw clinvar table
        self.HPApath = f"{self.raw_tables}HPA/proteinatlas.tsv"
        self.gencode_path = f"{self.raw_tables}gencode/GENEonly_cleaned_genecode_annotation.csv"  # Genecode Table

        self.processed_tables = f"{datadir}/processed_tables/"
        self.HGVSlookup_path = f"{self.processed_tables}HGVSlookup.csv" #Chrom to refID key table
        self.lastUpdate_file = f"{self.processed_tables}clinvar_lastUpdate.txt"

        #variables
        self.chroms = ['1', '2','3', '4', '5', '6', '7', '8', '9', '10', '11','12',
                         '13', '14', '15', '16', '17', '18', '19', '20', '21', '22','MT', 'Y', 'X']
        self.possible_queryTypes = ["coordinates","hgvs","phenotype"]
        self.possible_editors = ["all", "spCas9","baseeditor"]
    def process_refseq(self):
        """
        ftp : https://hgdownload.soe.ucsc.edu/goldenPath/hg38/database/ncbiRefSeq.txt.gz
        """
        mane = pd.read_csv(f'{self.processed_tables}MANE.GRch38.summary_cleaned.txt.gz')
        mane_ensid = list(mane['Ensembl_TranscriptID'])
        mane_tid = list(mane['TranscriptID'])

        labels = ['bin', 'id', 'chrom', 'strand', 'txStart', 'txEnd',
                  'cdsStart', 'cdsEnd', 'exonCount', 'exonStarts', 'exonEnds',
                  'score', 'name', 'cdsStartStat', 'cdsEndStat',
                  'exonFrames']
        out = gzip.open(f'{self.processed_tables}ncbiRefSeq.txt.gz', 'wt')
        for line in gzip.open(f'{self.raw_tables}Refseq/ncbiRefSeq.txt.gz', 'rt'):
            tokens = line.split('\t')
            if tokens[2].replace('chr', "") in self.chroms:
                try:
                    i = mane_tid.index(tokens[1])
                    tokens[0] = mane_ensid[i]
                except:
                    tokens[0] = '-'
                out.write('\t'.join(tokens[0:8] + tokens[9:11] + [tokens[12]] + [tokens[-1]]))
        out.close()

    def process_MANE(self):
        '''
        Mane has joins ENS and REFSEQ IDs
        '''
        mane = pd.read_csv(f'{self.raw_tables}clinvar/MANE.GRch38.v1.1.summary.txt.gz', sep="\t")

        mane['#NCBI_GeneID'] = mane['#NCBI_GeneID'].str.replace('GeneID:',"")
        mane = mane[mane['MANE_status'] == 'MANE Select']
        mane = mane.rename(columns = {'#NCBI_GeneID': 'GeneID','symbol':'GeneSymbol','RefSeq_nuc':'TranscriptID',
                             'RefSeq_prot':'ProteinID','chr_start':'Start',
                             'chr_end':'End','chr_strand':'Strand','GRCh38_chr':'ChrID','Ensembl_nuc':'Ensembl_TranscriptID'})

        mane = mane.drop(columns = 'MANE_status')
        mane = mane.loc[mane.ChrID.str.startswith('NC_')]
        mane.to_csv(f'{self.processed_tables}MANE.GRch38.summary_cleaned.txt.gz',index = False, compression='gzip')
    def process_clinvarVCF(self):
        '''
        clinvarvcf has molecular consequences
        '''
        mc = pd.read_csv('/groups/clinical/projects/editability/tables/raw_tables/clinvar/clinvar.vcf'
                         , comment='#', sep='\t', names=['Chr', 'PositionVCF', '?', 'REF', 'ALT', 'x', 'x1','attributes'])
        mc['chrHGVS_ID'] = mc['attributes'].str.extract(r'CLNHGVS=([^;]*)', expand=False)
        mc['MC'] = mc['attributes'].str.extract(r'MC=([^;]*)', expand=False)
        mc['AlleleID'] = mc['attributes'].str.extract(r'ALLELEID=([^;]*)', expand=False)
        con = []
        mc.loc[mc['MC'].isna(),'MC'] = '-'
        for x in mc['MC']:
            if '|' in x:
                x = ','.join([y for y in x.split('|') if 'SO' not in y])
            con.append(x)
        mc['MC'] = con

        mc = mc.drop(columns=['x1', 'x', 'attributes'])
        mc.to_csv(f'{self.processed_tables}clinvarvcf2txt.txt', index=False)

    def add_molecular_consequences(self):
        '''
        add molecular consequences
        '''
        for ch in self.chroms:
            df = pd.read_csv(f"{self.processed_tables}variant_tables/{ch}_variant.txt")
            mc = pd.read_csv(f'{self.processed_tables}clinvarvcf2txt.txt')
            mc['Chr'] = mc['Chr'].astype('str')
            mc = mc[mc['Chr'] == ch]
            mc = mc[['MC', 'AlleleID']]
            mc['AlleleID'] = mc['AlleleID'].astype('int64')
            joined = df.join(mc.set_index('AlleleID'), on='AlleleID')
            joined.to_csv(f'{self.processed_tables}variant_tables/{ch}_variant.txt', index=False)


    def extract_tid_from_hgvs(self,vdf):
        rprefix = r"((N(M|G|C|R)_[\d.]*)|(m))"
        rsuffix = r"(:(c|m|g|n)\.\S*)"
        hgs = list(vdf['HGVS_ID'])
        simple_ids = []
        tids = []
        for h in hgs:
            if re.search(rsuffix,h) and re.search(rprefix,h):
                tid = re.search(rprefix,h).groups()[0]
                suf = tid + re.search(rsuffix,h).groups()[0]
                tids.append(tid)
                simple_ids.append(suf)
            elif h.startswith('m'):
                tid = 'm'
                suf = h
                tids.append(tid)
                simple_ids.append(suf)

            else:
                print(h)
                tids.append('-')
                simple_ids.append('-')

        vdf.insert(3, 'HGVS_Simple', simple_ids)
        vdf.insert(4,'TranscriptID',tids)
        return vdf

    def add_MANE(self):
        mane = pd.read_csv(f'{self.processed_tables}MANE.GRch38.summary_cleaned.txt.gz')
        mane = mane[['Start', 'End', 'Strand', 'TranscriptID', 'ProteinID', 'Ensembl_TranscriptID', 'Ensembl_Gene']]

        for ch in self.chroms:
            vdf = pd.read_csv(f"{self.processed_tables}variant_tables/{ch}_variant.txt")
            mane['Ensembl_Gene'] = [x.split('.')[0] if type(x) == str else x for x in mane['Ensembl_Gene']]
            joined = vdf.join(mane.set_index('TranscriptID'), on = 'TranscriptID')

            joined.to_csv(f"{self.processed_tables}variant_tables/{ch}_variant.txt", index=False)
        #colreorder = [2,4,16,29,30,31,25,26,0,1,3,5,6,7,8,9,10,11,12,13,14,15,17,18,19,20,21,22,23,24,27,28,32,33,34]
        #joined = joined.iloc[:,colreorder]


    def extractOMIM(self,vdf):
        '''
        Find OMIM ID clinvar table and make seperate column
        '''
        all_ids = [",".join([x, y]) if len(",".join([x, y])) > 0 else "NA" for x, y in
                   zip(vdf['PhenoIDS'], vdf['OtherIDs'])]
        omim = []
        new_all_ids = []
        for x in all_ids:
            x = x.replace("MONDO:MONDO:", "MONDO:")
            found = list(set(re.findall("OMIM:([PS\.0-9]+)", x)))
            if len(found) == 0:
                omim.append("-")
            elif len(found) == 1:
                omim.append(found[0])
                x = x.replace(f"OMIM:{found[0]}", "")
            else:
                omim.append("|".join([z for z in found]))
                for z in found:
                    x = x.replace(f"OMIM:{z}", "")
            new_all_ids.append(x)
        vdf["OMIM"] = omim
        vdf["IDs"] = new_all_ids
        vdf = vdf.drop(columns=['PhenoIDS', 'OtherIDs'])
        return vdf

    def clean_clinvar(self):
        '''
        splits clinvar into files by chromosomes,
        removes unneeded columns
        Keeps only data from hg38 assembly
        '''
        ## Dropped cols
        # "LastEvaluated", "RS(dbSNP)", "Origin", 'Chromosome',
        # , "Cytogenetic", "ReviewStatus",
        # "NumberSubmitters", "Gudelines", "TestedInGTR", "SubmitterCategories"
        to_drop = ['RefAllele', 'AltAllele','Assembly', 'Start', 'Stop']
        in_file = gzip.open(self.clinvar_summary, "rt")
        contents = in_file.readlines()

        allcols = ['AlleleID', 'Type', 'HGVS_ID', 'GeneID', 'GeneSymbol', 'HGNC_ID',
                   'ClinicalSign', "ClinSigSimple", "LastEval", "RS#(dbSNP)", "nsv/esv (dbVar)",
                   "RCVaccession", "PhenoIDS","PhenoList", "Origin", "OriginSimple", "Assembly",
                   "ChrAccession", "Chr", "Start","Stop","RefAllele", "AltAllele",
                   "Cytogenetic", "ReviewStatus", "NumberSubmitters", "Guidelines", "TestedInGTR", "OtherIDs",
                   "SubmitterCategories", "VariationID", "PositionVCF", "RefAlleleVCF", "AltAlleleVCF"]

        cols = [allcols.index(c) for c in allcols if c not in to_drop]
        for ch in self.chroms:
            print(ch)
            out_fname = f"{self.processed_tables}/variant_tables/{ch}_variant.txt"
            lines = []
            for line in contents:
                line = line.split("\t")
                if line[18] == str(ch):
                    if line[16] != "GRCh37": # Remove hg19 data
                        line[-1] = line[-1].replace("\n", "")
                        if 'single nucleotide variant' in line:
                            line = [line[i] for i in cols]
                            lines.append(line)

            vdf = pd.DataFrame(lines, columns = [allcols[i] for i in cols])
            vdf['HGNC_ID'] = vdf['HGNC_ID'].apply(lambda x: x.replace("HGNC:",""))
            vdf['PositionVCF'] = vdf['PositionVCF'].astype('int')
            vdf = vdf[vdf['PositionVCF']>1]

            #keep only pathogenic
            vdf = vdf[vdf['ClinSigSimple'] == '1']##variants remaining = 12426
            vdf = vdf.sort_values('GeneSymbol')

            #Extract GeneID and GeneSymbol from HGVSID, In places that are missing
            blankgenes = vdf.loc[vdf['GeneID'] == '-1']
            hgvs = blankgenes['HGVS_ID']
            genenames = hgvs.str.extract(r"\((\w*\d*)\):").sort_values(0)
            temp = vdf.loc[~vdf['HGVS_ID'].isin(blankgenes)].drop_duplicates(subset='GeneSymbol')
            genenames = genenames.join(temp[['GeneID','GeneSymbol']].set_index('GeneSymbol'),on = 0)
            vdf.loc[vdf['HGVS_ID'].isin(hgvs),'GeneSymbol'] = genenames[0]
            vdf.loc[vdf['HGVS_ID'].isin(hgvs),'GeneID'] = genenames['GeneID']
            vdf.loc[vdf.GeneID.isna(),'GeneID'] = -1
            vdf['GeneID'] = vdf['GeneID'].astype('int64')
            vdf = vdf.reset_index(drop=True)

            # Find and extract OMIM ID
            vdf = self.extract_tid_from_hgvs(vdf)
            vdf = self.extractOMIM(vdf)
            vdf.to_csv(out_fname,index=False)

        self.add_MANE()
        self.add_molecular_consequences()


    def intervalMatch(self,df1,df2):
        '''
        matches a snp position to gene coord by creating dummy indexes
        '''
        df1 = df1.reset_index() #clinvar
        df2 = df2.reset_index()  # hpa
        temp1 = [f'unmatched{str(x)}' for x in df1["PositionVCF"].index] #dummy clin var index
        temp2 = [f'matched{x}' for x in df2["Position"].index] #dummy hpa index
        starts = [int(x.split("-")[0]) for x in df2.Position]
        ends = [x.split("-")[1] for x in df2.Position]
        for i in df1.index:
            posvcf = df1["PositionVCF"].iloc[i]
            new_end = [y for x, y in zip(starts,ends) if int(posvcf) > int(x)]
            new_p2 = [y for y in new_end if int(posvcf) < int(y)]
            if len(new_p2)>0:
                # set clin var dummy interval to match hpa
                temp1[i] = f"matched{ends.index(new_p2[0])}"
        return temp1,temp2

    def make_gene_tables(self):
        '''
        attached HPA gene expression info to clinvar info by ensembl id
        '''

        hpa = pd.read_csv(self.HPApath, delimiter='\t')

        cols = ['Ensembl', 'Chromosome', 'Gene description', 'Protein class', 'Biological process',
                'Molecular function', 'Uniprot',
                'Disease involvement', 'RNA tissue specificity', 'RNA tissue specific nTPM',
                'RNA tissue distribution', 'RNA tissue cell type enrichment',
                'RNA single cell type specific nTPM', 'RNA tissue specific nTPM']
        hpa = hpa[cols]

        mane = pd.read_csv(f'{self.processed_tables}MANE.GRch38.summary_cleaned.txt.gz')
        mane = mane[['TranscriptID','Start', 'End', 'Strand',  'ProteinID', 'Ensembl_TranscriptID', 'Ensembl_Gene']]
        mane['Ensembl_Gene'] = [x.split('.')[0] if type(x) == str else x for x in mane['Ensembl_Gene']]
        joined_df = hpa.join(mane.set_index('Ensembl_Gene'), on='Ensembl_Gene')
        joined_df.to_csv(f'{self.processed_tables}gene_tables.csv.gz', index=False,compression='gzip')

    def make_HGVStable(self):
        '''
        Create CSV file of unduplicated HGVS prefix(coding ref name) and Chromosome
        '''
        names, chrs = [], []
        for ch in self.chroms:
            clin = pd.read_csv(f"{self.processed_tables}variant_tables/{ch}_variant.txt")
            names += list(set(clin['TranscriptID']))
            chrs += [ch for i in range(len(set(clin['TranscriptID'])))]

        df = pd.DataFrame({"TranscriptID": names, "Chr": chrs}).drop_duplicates()
        out_file = f"{self.processed_tables}HGVSlookup.csv"
        df.to_csv(out_file, index=None)

    def updateTables(self):
        # TODO: add user inquiry to whether they want to init update
        print("updating to latest version of clinvar")

        cmd = f"wget {self.clinvar_ftp} -O {self.clinvar_summary}"
        p = subprocess.run(cmd, shell=True,
                           capture_output=True)
        print(p)

        cmd = f"wget {self.refseq} -O {self.raw_tables}Refseq/ncbiRefSeq.txt.gz"
        p = subprocess.run(cmd, shell=True,
                           capture_output=True)
        print(p)

        cmd = f"wget https://ftp.ncbi.nlm.nih.gov/refseq/MANE/MANE_human/current/MANE.GRCh38.v1.1.summary.txt.gz -O " \
              f"{self.raw_tables}clinvar/MANE.GRch38.v1.1.summary.txt.gz"
        p = subprocess.run(cmd, shell=True,
                           capture_output=True)
        print(p)

        cmd = f"wget {self.clinvar_vcf} -O {self.raw_tables}clinvar/clinvar.vcf.gz"
        p = subprocess.run(cmd, shell=True,
                           capture_output=True)
        print(p)
        cmd = f"wget {self.clinvar_index} -O {self.raw_tables}clinvar/clinvar.vcf.gz.tbi"
        p = subprocess.run(cmd, shell=True,
                           capture_output=True)

        cmd = f"bcftools view -f type!=snp {self.clinvar_vcf} -o {self.raw_tables}clinvar/clinvar.vcf -O v"
        p = subprocess.run(cmd, shell=True,
                           capture_output=True)

        print("Cleaning and Splitting Clinvar.....")
        self.clean_clinvar()
        print("Adding Ensembl Identifiers....")
        self.add_MANE()
        print("Adding Molecular Consequences....")
        self.add_molecular_consequences()
        print("Appending HPA data.....")
        self.appendHPA()
        print("Writing new HGVS Lookup table.....")
        self.make_HGVStable()

        with open(self.lastUpdate_file, "w") as f:
            today = date.today()
            f.write(str(today))
        f.close()

#     def check_updates(self):
#         """
#         Determines if an update is needed based on checking the date of txt file
#         '''
#
#         f = open(self.lastUpdate_file, "r").readlines()
#         lastdate = datetime.strptime(f[0], '%Y-%m-%d').date()
#         if (date.today() - lastdate).days > 31:
#             self.updateTables()
#         else:
#             print('You are using the latest clinvar data')
#             print(f"Last updated {lastdate}")
#
#     def validate_input_file(self, input_file):
#         cols = ["Query", "Type","Editor"]
#         try:
#             self.input_df = pd.read_csv(input_file,usecols= cols)
#         except:
#             raise FileNotFoundError(f"There was was a problem reading {input_file}. \n"
#                                     f"Be sure this is a csv file with the column headers {cols}")
#
#         for term in self.input_df['Type'].unique():
#             if term not in self.possible_queryTypes:
#                 raise ValueError(f"Query type must be either {self.possible_queryTypes}")
#
#         for term in self.input_df['Editor'].unique():
#             if term not in self.possible_editors:
#                 raise ValueError(f"Editor type must be either {self.possible_editors}")
#         return self.input_df
#
#
# '''
# datadir = "/groups/clinical/projects/editability/tables/"
# val = Validator(datadir)
# val.clean_clinvar()
# val.add_MC(
#
# val.make_guidetabs()
#
# for ch in val.chroms:
#     vdf = pd.read_csv(f"{val.processed_tables}{ch}_variant.txt")
# '''


import gzip
from Bio import SeqIO, SeqUtils
from Bio.Seq import Seq


def get_refseq_entry(term, field, annote_path):
    '''
    Using ncbiRefSeq.txt to find cds features by either interval, gene name or transcript ID
    example input:
    term, field = 'NM_000532.5', 'tid'
    term, field = 'ENST00000251654.9', 'eid'
    term, field = 'PCCB','name'
    term,field =  'chr3:136250339-136330169','interval'
    '''

    labels = ['eid', 'tid', 'chrom', 'strand', 'txStart', 'txEnd',
              'cdsStart', 'cdsEnd', 'exonStarts', 'exonEnds', 'name', 'exonFrames']

    if field != 'interval':
        not_found = True
        for line in gzip.open(annote_path, 'rt'):
            tokens = line.split('\t')
            entry = dict(zip(labels, tokens))
            if term in entry[field]:
                not_found = False
                break

        if not_found:
            entry = None
            print(f"{term} not found in refseq data")

            if '.' in term:
                new_term = term.split('.')[0]
                print(f'searching for {new_term} instead')
                entry = get_refseq_entry(new_term, field,annote_path)

    else:  # only used for intervals search
        not_found = True
        ch = term.split(":")[0]
        start, end = term.split(":")[1].split('-')
        pos = int((int(start) + int(end)) / 2)

        for line in gzip.open(annote_path, 'rt'):
            tokens = line.split('\t')
            entry = dict(zip(labels, tokens))
            if ch == entry['chrom']:
                if pos in range(int(entry['txStart']), int(entry['txEnd'])):
                    not_found = False
                    break
        if not_found:
            entry = None
            print(f"{term} not found in refseq data")

    return entry


def get_cds_info(tx_seq, entry):
    '''
    uses entry info to find cds (without utr's)
    '''
    exon_starts = entry['exonStarts'][:-1].split(',')
    exon_ends = entry['exonEnds'][:-1].split(',')
    exon_frames = entry['exonFrames'].replace("\n", "")[:-1].split(',')
    tx_start = int(entry['txStart'])

    exons = [(int(exon_starts[i]) - tx_start, int(exon_ends[i]) - tx_start) for i in range(len(exon_ends))]
    for i in range(len(exon_frames)):
        if exon_frames[i] == '-1':  # -1 means entire exon is UTR
            exons = exons[1:]
            exon_starts = exon_starts[1:]
        else:
            break
    for i in range(1, len(exon_frames)):
        if exon_frames[-i] == '-1':
            exons = exons[0:len(exons) - 1]
            exon_ends = exon_ends[0:-1]
        else:
            break

    # Determine the stop and start of UTR
    if len(exons) > 0:
        exons[0] = (int(entry['cdsStart']) - int(exon_starts[0]) + exons[0][0], exons[0][1])
        exons[-1] = (exons[-1][0], exons[-1][1] - (int(exon_ends[-1]) - int(entry['cdsEnd'])))

        cds = Seq(''.join([str(tx_seq)[a:b] for a, b in exons]))
        if entry['strand'] == '-':
            cds = cds.reverse_complement()
    else:
        cds = None
        exons = None

    # translation = cds.translate()
    return [exons, tx_seq, cds]

def find_transcript_info(term, fasta, annote_path):
    '''
    Using a Refseq Transcript_ID, Ensembl Transcript_ID or coordinates find transcript annotations and transcript sequence
    from either a genome fasta path or given genome sequence
    '''
    # id= 'NM_000532.5' or 'ENST00000251654.9'
    # fasta = f"/groups/clinical/projects/clinical_shared_data/hg38/hg38_chr20.fa.gz"
    if type(fasta) == str:
        fasta_seq = SeqIO.read(gzip.open(fasta, 'rt'), 'fasta')
    else:
        fasta_seq = fasta
    field = 'eid' if term.startswith('E') else 'tid' if term.startswith('N') else 'interval'

    entry = get_refseq_entry(term=term, field=field, annote_path=annote_path)
    if entry != None:
        tx_seq = fasta_seq.seq[int(entry['txStart']):int(entry['txEnd'])]
        tid_info = get_cds_info(tx_seq, entry)
    else:
        tid_info = None
    return entry, tid_info


def find_codons(dist_from_cds_start):
    '''
    Finds reading frame of SNV in extracted sequence
    '''
    rf = 1 if dist_from_cds_start % 3 == 2 else 2 if dist_from_cds_start % 3 == 0 else 0
    return rf


def find_snvseq_info(seq,snvpos,exons, cds, entry,window):
    # returns - sequence,feature,translation(if needed)
    # feature: non-coding, utr5,ut3,intron,exon, start_codon, stop_codon
    # snvpos, alt = 11576257, 'T'
    global dist_from_cds_start
    feature, rf = None, None

    t_snvpos = int(snvpos) - int(entry['txStart'])
    cdstart, cdsend = int(entry['cdsStart']) - int(entry['txStart']), int(entry['txEnd']) - int(entry['txStart'])
    strand = entry['strand']

    if t_snvpos < 0 or cds == None:
        # not in transcript - shouldn't happen or else no entry would be found
        feature = 'non-coding'

    else:
        if t_snvpos in range(cdstart, cdsend + 1):
            # in CDS
            # find if utr
            feature = 'intron'
            if t_snvpos in range(cdstart, exons[0][0] + 1):
                feature = '3utr' if strand == '-' else '5utr'
            elif t_snvpos in range(exons[-1][1], cdsend + 1):
                feature = '5utr' if strand == '-' else '3utr'

            # find if exon or intron
            else:
                exon_n = 0
                for x in exons:
                    # if in exon find reading frame
                    if t_snvpos in range(x[0], x[1] + 1):
                        # stop and start codon
                        feature = 'exon'
                        dist = sum([e[1] - e[0] for e in exons[0:exon_n]])
                        dist_from_cds_start = dist + (t_snvpos - x[0])

                        if strand == '-':
                            dist_from_cds_start = (len(cds) - dist_from_cds_start) + 1

                        if dist_from_cds_start < 3:
                            feature = 'start_codon'

                        if dist_from_cds_start > len(cds) - 3:
                            feature = 'stop_codon'

                        rf = find_codons(dist_from_cds_start)
                        break
                    exon_n += 1
        else:
            # in transcript but not in cds
            seq = str(seq)

            if len(SeqUtils.nt_search(seq[window - 6:window + 5], 'TTTATT')) > 1 or len(
                    SeqUtils.nt_search(seq[window - 6:window + 5], 'AATAAA')) > 1:
                feature = 'polya'
            elif len(SeqUtils.nt_search(seq[window - 6:window + 5], 'TATAAA')) > 1 or len(
                    SeqUtils.nt_search(seq[window - 6:window + 5], 'ATATTT')) > 1:
                feature = 'promoter'
            elif len(SeqUtils.nt_search(seq[window - 7:], 'GGNCAATCT')) > 1:
                if len(SeqUtils.nt_search(seq[window - 7:window + 6], 'GGNCAATCT')) > 1:
                    feature = 'promoter'
                else:
                    feature = 'TSS'
            elif len(SeqUtils.nt_search(seq[window + 6:], 'AGATTGNCC')) > 1:
                if len(SeqUtils.nt_search(seq[window - 7: window + 6], 'AGATTGNCC')) > 1:
                    feature = 'promoter'
                else:
                    feature = 'TSS'
            else:
                feature = 'flanking'

    return feature, rf




# Native modules
import os
import re
from pathlib import Path
# Installed modules
import yaml
import gdown
import pandas as pd
import psycopg2
from psycopg2 import errors
from sqlalchemy import create_engine, text
from sqlalchemy.schema import CreateSchema as cschema


def get_absolute_path(name):
	for root, dirs, files in os.walk(Path.home()):
		for item in dirs + files:
			if re.search(fr"\b{name}\b", item):
				return os.path.abspath(os.path.join(root, item))


def psql_connect(config_handle):
	database_name = config_handle['database_name']
	# Uses the parameters in the YAML config_render file to setup the DB connection string
	conn_string = 'postgresql://{}:"{}"@{}:{}/{}'.format(
		config_handle['username'],
		# This is how the password will be defined
		# os.environ['PSQL_PASS'],
		# This is a placeholder until this code lives in its long-term server
		config_handle['passwd'],
		config_handle['hostname'],
		config_handle['port'],
		database_name
	)

	return conn_string


def get_col_as_list(csv_path, target_col):
	list_of_links = pd.read_csv(csv_path)[target_col].tolist()
	return list_of_links


def get_excel_from_google_drive(url_list, id_list, start_tab_index=1):
	association_dict = {}

	for idx in range(len(url_list)):
		sheets_dict = {}
		url = url_list[idx]
		output_dir = get_absolute_path("jobs")
		output_file = f'{output_dir}temp.xlsx'

		gdown.download(url, output_file, quiet=False, fuzzy=True)  # Download the file using gdown
		xl = pd.read_excel(output_file, sheet_name=0, engine='openpyxl')
		sheet_names = list(xl.keys())[start_tab_index:]

		for sheet_name in sheet_names:
			cut_sheet_name = re.sub('_table', '', sheet_name, flags=re.IGNORECASE)
			sheet_split = re.split('_', cut_sheet_name)
			sheet_name_loop = ''
			for field in range(0, 3):
				try:
					sheet_name_loop += f"{sheet_split[field]}_"
				except IndexError:
					break
			clean_sheet_name = sheet_name_loop.strip('_')

			# Load the Excel file into a pandas DataFrame
			df = pd.read_excel(output_file, sheet_name=sheet_name)
			sheets_dict.setdefault(clean_sheet_name, df)

		# Remove Temps
		os.remove(output_file)

		clean_id = re.sub(r'(\S+) \S+', r'\1', id_list[idx])
		print(clean_id)
		association_dict[clean_id] = sheets_dict
	# Return the dictionary
	return association_dict


def load_wiki_tables2db(association_dict, conn_string, schema_prefix):
	print("Create DB engine")
	db_engine = create_engine(conn_string, echo=True)
	db_connection = db_engine.connect()
	print("Connection established with the DB")
	for unique_id in association_dict:
		schema_name = f"{schema_prefix}.{unique_id}"
		for table_name in association_dict[unique_id]:
			table = association_dict[unique_id][table_name]
			print("Looping through tables and adding to {} schema".format(schema_name))
			try:
				if not db_connection.dialect.has_schema(db_connection, schema_name):
					db_connection.execute(cschema(schema_name))

				# Guess data types before loading the df into DB
				ready_table = table.convert_dtypes().infer_objects()
				ready_table.update(ready_table.select_dtypes('object').astype(str))

				# Load df into the relevant schema
				ready_table.to_sql(table_name, db_engine,
				                      schema=schema_name,
				                      if_exists='replace',
				                      index=False)

			except psycopg2.errors.InFailedSqlTransaction:
				print("Could not add table to schema")
				return
	db_connection = psycopg2.connect(conn_string)
	db_connection.commit()
	db_connection.close()
	print("Commited changes to DB")


# Load config_render file
with open("config/caspedia_ingest.yaml", "r") as f:
	config = yaml.safe_load(f)


def main():
	# Get schema name from the config file
	schema = config["schema"]
	# Set up the DB connection string using parameters from the config file
	conn_string = psql_connect(config)

	# Consolidate Cloud links and identifiers
	wiki_links_list = get_col_as_list(config["manifest_path"], config["links_column"])
	uniq_id_list = get_col_as_list(config["manifest_path"], config["unique_id_col"])

	# Download forms from Google-drive
	id_to_sheets_dict = get_excel_from_google_drive(wiki_links_list, uniq_id_list)

	load_wiki_tables2db(id_to_sheets_dict, conn_string, schema)


if __name__ == "__main__":
	main()



# Native Modules
import subprocess
# import regex as re
# import sys
import os
from datetime import date  # datetime, date
# Installed Modules
import pandas as pd
from Bio.SeqRecord import SeqRecord
from Bio.SeqIO import FastaIO
from Bio.Blast.Applications import NcbiblastnCommandline as blastn
from Bio.Blast.Applications import NcbimakeblastdbCommandline as makedb
from Bio.Blast import NCBIXML
# from Bio import motifs


def set_export(outdir):
	# Create outdir if inexistent
	if not os.path.exists(outdir):
		os.makedirs(outdir)
	return outdir


def blast_id_parsing(id_string, sep_dict):
	loop_id = id_string
	for level in sep_dict:
		try:
			loop_id = loop_id.split(sep_dict[level][0])[sep_dict[level][1]]
		except IndexError:
			continue
	return loop_id


def hit_report_limit(query_to_hit_dict, nkeep):
	filtered_dict = {}
	hit_id_list = []
	# Within each blast query, keep a maximum of nkeep hits based on lowest evaue
	for query in query_to_hit_dict:
		series = pd.DataFrame.from_dict(query_to_hit_dict[query], orient='index')
		fseries = series['e-value'].nsmallest(nkeep, keep='first')
		for hit in fseries.index.tolist():
			# Keep the same structure of the input dictionary
			filtered_dict.setdefault(query, {}).setdefault(hit, query_to_hit_dict[query][hit])
			hit_id_list.append(hit)
	return filtered_dict, hit_id_list


def blast_result_parser(xml_temp_path, eval_threshold, sep_instructions, nkeep):
	report_dict = {}
	for record in NCBIXML.parse(open(xml_temp_path)):
		if record.alignments:
			for align in record.alignments:
				for hsp in align.hsps:
					if hsp.expect < eval_threshold:
						id = blast_id_parsing(align.hit_def, sep_instructions)
						report_dict.setdefault(record.query, {}).setdefault(id, {
							"blastp_alignment_len": align.length, "e-value": hsp.expect})
	filtered_report_dict, hit_id_list = hit_report_limit(report_dict, nkeep)
	return filtered_report_dict, hit_id_list


def main():
	# SNAKEMAKE IMPORTS
	#   Inputs
	# The aim is to have a multifasta containing all the syntenic
	# regions in the pangenomes to the canonical region in the Hg38 assembly
	# It's crucial that the path where these are stored is isolated for the creation of a BlastDB
	pangenome_multifasta = str(snakemake.input.multifasta)
	# This will be the consensus sequence between patient's VCF and the Hg38 assembly
	consensus_fasta_path = str(snakemake.input.query_fasta)
	#   Outputs
	resultsfolder = set_export(str(snakemake.output))
	#   Params
	db_path = str(snakemake.params.db_path)
	evalue = str(snakemake.params.blast_evalue_thresh)
	threads = int(snakemake.wildcards.threads)
	blastout_path = str(snakemake.params.blastout_path)

	# Internal path variables
	temp = f"{blastout_path}/temp.xml"

	# MakeblastDB
	makedb_cline = makedb(dbtype="prot",
	                      input_file=pangenome_multifasta)
	makedb_cline()
	# Blast run
	print("BlastP Run in progress")
	blast_cline = blastn(query=consensus_fasta_path,
	                     db=db_path,
	                     task='megablast',
	                     num_threads=threads,
	                     evalue=evalue,
	                     out=temp,
	                     outfmt=5)

	print(f"Blast command used:\n {blast_cline}")
	blast_cline()

	# Format parser
	# TODO: Adjust FASTA ID separator and the number of hits to keep
	print("Parsing BlastP output")
	hit_dict, hit_list = blast_result_parser(temp,
	                                         evalue,
	                                         config["id_sep_dict"],
	                                         config["blast_nkeep"])
	# Remove blast XML temp file
	os.remove(temp)


if __name__ == "__main__":
	main()



# == Native Modules

# == Installed Modules
from setuptools import setup, find_packages
# == Project Modules


setup(
    name='meditability',
    version='0.1.2',
    description='',
    author='Interventional Genomics Unit',
    author_email='',
    entry_points={
        "console_scripts": [
            "medit = prog:main"
        ]
    },
    packages=find_packages(),
    install_requires=[
        'snakemake>=7.32.4',
        'biopython>=1.81',
        'pyyaml>=6.0',
        'pytz>=2023.3'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
    ],
)



# Native Modules
import gzip
import zlib

# import regex as re
import os
import re
from zlib import error
# Installed Modules
import pandas as pd
from Bio import SeqIO, SeqUtils
from Bio.Seq import Seq
import pickle
# Project Modules
from dataH import DataHandler
from annotate import get_refseq_entry, find_snvseq_info, find_transcript_info

###############
# Main Script with Fetch_Guides Class for running pipeline
###############


def set_export(outdir):
	# Create outdir if inexistent
	if not os.path.exists(outdir):
		os.makedirs(outdir)
	return outdir


class Fetch_Guides:

	def __init__(self,
	             queries: list,
	             qtype: str,
	             editor_request: str | list,
	             be_request: str | list,
	             editors: dict,
	             base_editors: dict,
	             datadir: str,
	             fasta_path: str,
	             annote_path: str,
	             **kwargs):
		"""
		:param queries: list of query terms, either in hgvs format - 'NM_000518.5:c.114G>A' or coords 'chr11:5226778C>T' (COORDS ALLELES MUST BE PLUS STRAND!!)
		:param qtype: 'hgvs' or 'coord'
		  --> if 'hgvs', providing the coordinates in the kwargs with 'hgvscoord' can reduce processing time
		  --> hgvs assumes the query is already in clinvar and will generate a variant report with the gene report,
		  --> if 'coord' then just gene report is created
		:param editor_request: 'clinical', 'custom', name (list/str from editor choices)
		--> custom must contain kwargs - pam, pamISFirst,guidelen (optional:name,win_size)
		:param be_request: 'off','default','all', or select BE editor for base editor choices below
		:param editors: Dictionary containing information on the current set of editors supported by mEdit
		:param base_editors: Dictionary containing information on the current set of base editors supported by mEdit
		:param genome: genome used
		:param datadir: folder where tables and pre-computed data live
		:param fasta_path: *Unsure using chromsome seperate files right now but unsure if this will be permenant
		:param kwargs: 'hgvscoord' , 'clin_report','gene_report'

		"""
		##-----------------User Inputs--------------------##
		if qtype == 'hgvs':
			self.queries = self.validate_hgvs(queries)
		if qtype == 'coord':
			self.queries = self.validate_coord(queries)
		self.qtype = qtype
		self.editor_request = editor_request
		self.be_request = be_request
		self.hgvscoord = None
		self.clin_report = True
		self.gene_report = True
		self.kwargs = kwargs
		self.editor_lib = editors
		self.be_lib = base_editors

		if 'hgvscoord' in kwargs.keys():
			self.hgvscoord = self.validate_coord(kwargs['hgvscoord'])  # 'chr11:5226778C>T'
		if 'gene_report' in kwargs.keys():
			self.gene_report = kwargs['gene_report']
		if 'clin_report' in kwargs.keys():
			self.clin_report = kwargs['clin_report']

		# input paths/folders
		self.processed_tables = f"{datadir}/processed_tables"  # folder with cleaned clinvar/hpa tabs
		self.HGVSlookup_path = f"{self.processed_tables}/HGVSlookup.csv"
		self.fasta_path = fasta_path
		self.annote_path = annote_path

		# other variables
		self.snv_info = {}  # {chrom: (id,snv_pos,ref,alt)}

		##---------------libraries and keys--------------------##
		self.editor_choices = list(editors.keys())
		self.BE_choices = list(base_editors.keys())

		## ------------Defaults and settings------------------##
		##configure editor options
		self.search_params = self.configure_search_params()

		# configure BE options
		if self.be_request != 'off':
			self.BE_search_params = self.set_BE_params()
		# ---------------Flags--------------------------#
		self.clininfo_flag = False

		# ---------------Ouputs--------------------------#
		self.all_variant = pd.DataFrame()
		self.all_gene = pd.DataFrame()
		self.all_guides = {}
		self.all_BE = {}


	def configure_search_params(self):
		"""
		set paramteres for the selected editor or editors(not BE editors)
		"""
		# search for all guides
		if self.editor_request == 'clinical':
			search_params = self.editor_lib['clinical']
		# set custom editor params
		elif self.editor_request == 'custom':
			search_params = self.set_params(self.kwargs)

		# search for selected subset
		# TODO: The guide_prediction.py needs to ingest this correctly
		elif type(self.editor_request) is list:
			search_params = {}
			for e in self.editor_request:
				search_params[e] = self.editor_lib[e]

		# else use single set parameters
		else:
			if self.editor_request in self.editor_choices:
				search_params = {self.editor_request: self.editor_lib[self.editor_request]}

			else:
				print('Please choose a name(s) found in the editor name choices')

		print(f'Editor(s) set to: {[x for x in search_params.keys()]}')
		return search_params

	def set_params(self, kwargs):
		name = 'custom'

		try:
			pam = kwargs['pam']
			pamISfirst = kwargs['pamISfirst']
			guidelen = kwargs['guidelen']
			if pamISfirst == False:
				win_size = -2
			else:
				win_size = guidelen - 2
		except KeyError:
			print("custom editor selection MUST include a minimum of kwargs = pam, pamISFirst,guidelen")

		if 'name' in kwargs.keys():
			name = kwargs['name']
		if 'win_size' in kwargs.keys():
			win_size = kwargs['win_size']

		params = {name: (pam, pamISfirst, win_size, guidelen, '')}

		return params

	def set_BE_params(self):
		# sets base editor search params, each key is a list of 2 or more; refernce seq search params,
		# then any set that follows starts with the conversion (ex. 'AG' is A --> G) and then the base editors that have the same params

		if self.be_request == 'default':
			self.BE_search_params = self.be_lib['default']
		elif self.be_request == 'all':
			self.BE_search_params = self.be_lib['all']

		# TODO: Setup BE user-defined list processing
		elif type(self.be_request) is list:
			self.BE_search_params = {}
			for e in self.be_request:
				self.BE_search_params[e] = self.be_lib['all'][e]

		else:
			if self.be_request not in self.BE_choices:
				print('That is not a valid Base Editor')
				print(f'please choose from {self.BE_choices}')
			else:
				for k, v in self.be_lib['all'].values():
					if self.be_request in v[1][-1]:
						self.BE_search_params = {self.be_request: self.be_lib['all'][k][0:2]}

					if len(v) == 3:
						if self.be_request in v[2][-1]:
							self.BE_search_params = {self.be_request: self.be_lib['all'][k][0] + self.be_lib['all'][k][2]}

		return self.BE_search_params

	def write_gsearch_params(self, outfile):
		# writes pickle of selected guide search params for later use in process_genome
		# 'editor', 'pam', '5prime_pam','guide_length', 'DSB site', 'notes'
		with open(outfile, 'ab') as gfile:
			pickle.dump(self.search_params, gfile)


	def write_snv_site_info(self, outfile):
		'''
		#writes pickle of SNV site info for later use in process genome
		#query, tid, eid, strand, ref, alt, feature_annotation, extracted_seq, codons, coord
		'''
		with open(outfile, 'ab') as sfile:
			pickle.dump(self.snv_info, sfile)

	def write_guide_csv(self, guides, outfile):
		df = pd.DataFrame(guides)
		if 'On-Target Efficiency Score' in df.columns:
			temp = df[df['On-Target Efficiency Score'] != '-'].sort_values(by='On-Target Efficiency Score', ascending=False)
			df = pd.concat([temp, df[df['On-Target Efficiency Score'] == '-']]).reset_index(drop=True)
		df['Guide_ID'] = [y + str(x) for x, y in zip(list(df.index), list(df['Guide_ID']))]
		df.to_csv(outfile, index=False)
		return df

	def add_clininfo(self, gene_out, variant_out):
		if not self.clininfo_flag:
			# TAYLOR: Here we can probably provide something more informative.
			#   Keeping it as a placeholder
			self.all_gene.to_csv(gene_out)
			print("GENES AND VARIANT TABLES ARE UNAVAILABLE")
			return
		all_tids = []
		for ch, data in self.snv_info.items():
			all_tids += [d[1] for d in data]

			if self.qtype == 'hgvs':
				tempvar = pd.read_csv(f"{self.processed_tables}/variant_tables/{ch}_variant.txt")
				tempvar = tempvar.loc[tempvar['HGVS_Simple'].isin(list(self.queries))]
				self.all_variant = pd.concat([self.all_variant, tempvar])
		#TODO: The path to gene_tables will ideally be imported as a snakemake variable
		tempgene = pd.read_csv(f"{self.processed_tables}/gene_tables/gene_tables.csv")
		self.all_gene = tempgene.loc[tempgene['TranscriptID'].isin(list(all_tids))]

		# datenow = date.today().strftime('%Y-%m-%d')

		# gene_out = f"{self.resultsfolder}/Gene_Report.csv"
		print(f"\n READY TO PRINT GENE OUT TO: {gene_out}\n ")
		self.all_gene.to_csv(gene_out, index=False)

		if self.qtype == 'hgvs':
			# variant_out = f"{self.resultsfolder}/Variant_Report.csv"
			print(f"\nREADY TO PRINT VARIANT OUT TO: {variant_out}\n")
			self.all_variant.to_csv(variant_out, index=False)

	@staticmethod
	def extract_seqs(searchseq, pos, alt, window):
		"""
		extracts the sequence +/-30bp surrounding a SNV then swaps ref for alt allele
		"""
		extracted_seq = str(searchseq[pos - window:pos + window])
		extracted_seq = Seq(extracted_seq[0:window] + alt + extracted_seq[window + 1:]).upper()
		return extracted_seq

	def get_chroms(self,queries):
		hgvs_tab = pd.read_csv(self.HGVSlookup_path)
		q_prefixes = [x.split(':')[0] for x in queries]
		chroms = set(hgvs_tab.loc[hgvs_tab['TranscriptID'].isin(q_prefixes), 'Chr'])
		return chroms

	def fetch_query_info(self):
		# Gets Transcript info
		global term
		snv_info = {}
		window = 50

		# If quering by HGVSID with no other info then need to get chromsome/location/alt/ref
		if self.qtype == 'hgvs' and self.hgvscoord is None:
			print("Looking up HGVS in Clinvar.......")
			chroms = self.get_chroms(self.queries)

			for ch in chroms:
				df = pd.read_csv(f"{self.processed_tables}/variant_tables/{ch}_variant.txt", low_memory=False)
				gadf = df.loc[df['HGVS_Simple'].isin(self.queries)]
				snv_info[ch] = gadf[['HGVS_Simple', 'PositionVCF', 'RefAlleleVCF', 'AltAlleleVCF']].to_dict('tight')[
					'data']

		# Else All information is given to find transcript info
		else:
			coords = self.queries if self.qtype == 'coord' else self.hgvscoord

			coord_fmt = r'chr[0-9MTXY]*:(\d*)([ATCG]{1})\>([ATCG]{1})'

			print(f"\n BUG INSPECTION \n Coords: {coords}\n Queries: {self.queries}\n")
			print(f"PREMISSAS:\n Qtype: {self.qtype}\n Hgvs Coord: {self.hgvscoord}")
			for x in range(len(self.queries)):
				print(f" Current Query: {x}\n")
				ch = coords[x].split(':')[0].replace('chr', '')
				if ch not in snv_info.keys():
					snv_info[ch] = []
				snvpos, alt, ref = list(re.search(coord_fmt, coords[x]).groups())
				snv_info[ch].append([self.queries[x], int(snvpos), alt, ref])

		self.snv_info = snv_info

		print("Gathering Variant Genomic Info.......")

		for ch, data in snv_info.items():  # find transcript info
			# === Transitioning SeqIO.read to direct import of Pickled SeqRecord Objects ===
			# chr_fasta_path = self.fasta_path.replace('.fa.gz', f'_chr{str(ch)}.fa.gz')
			chr_fasta_path = f"{self.fasta_path}/chr{str(ch)}.pkl"
			try:
				print(f"Finding transcripts information: Assessing {chr_fasta_path}")
				# fasta = SeqIO.read(gzip.open(chr_fasta_path, 'rt'), 'fasta')
				with open(chr_fasta_path, 'rb') as pfile:
					fasta = pickle.load(pfile)
				new_data = []
			except FileNotFoundError:
				print(f"The file {chr_fasta_path} was not found. Please regenerate background data and check the target directory")
				continue
			except pickle.UnpicklingError:
				print(f"The file {chr_fasta_path} is not in the correct format. Please regenerate background data")
				continue

			for d in data:
				query, snvpos, ref, alt = d
				print(query, ref, alt, snvpos)
				if self.qtype == 'hgvs':  # pull refseqID from HGVS and search transcript by this
					term = query.split(':')[0]
				if self.qtype == 'coord':  # else use coordsinates to search trancript
					term = f"chr{str(ch)}:{str(snvpos)}-{str(snvpos)}"
				entry, tid_info = find_transcript_info(term=term, fasta=fasta, annote_path = self.annote_path)

				if entry is not None:
					exons, tx_seq, cds = tid_info
					t_snvpos = int(snvpos) - int(entry['txStart'])
					extracted_seq = self.extract_seqs(searchseq=tx_seq, pos=t_snvpos - 1, alt=alt, window=window)
					if len(extracted_seq) != window * 2:
						# if flanking or utr the extracted seq needs to come from the chromosome file
						extracted_seq = self.extract_seqs(searchseq=fasta.seq[snvpos - 100:snvpos + 100], pos=t_snvpos - 1,
												alt=alt,
												window=window)

					feature_annotation, codons = find_snvseq_info(extracted_seq,snvpos,exons, cds, entry,window)
					strand = entry['strand']
					tid, eid = entry['tid'], entry['eid']

				else:
					feature_annotation = 'undetermined/non-coding'
					codons = 'None'
					strand = '+'
					extracted_seq = self.extract_seqs(searchseq=fasta.seq, pos=snvpos, alt=alt, window=window)
					tid, eid = term, '-'
				new_data.append(
					[query, tid, eid, strand, ref, alt, feature_annotation, extracted_seq, codons,
					 f"chr{str(ch)}:{str(snvpos)}"])

				print('Query term & annoation:', query, feature_annotation)
			snv_info[ch] = new_data

		self.snv_info = snv_info

	@staticmethod
	def validate_hgvs(queries):
		'''
		standardizes input hgvs and checks formating
		'''
		rprefix = r"((N(M|G|C|R)_[\d.]*)|(m))"
		rsuffix = r"(:(c|m|g|n)\S*)"
		validated_queries = []
		for q in set(queries):
			if re.search(rsuffix, q) and re.search(rprefix, q):
				validated_queries.append(re.search(rprefix, q).groups()[0] + re.search(rsuffix, q).groups()[0])
			else:
				print(q)
		n = len(validated_queries)
		print(f'{n} out of {len(queries)} HGVS IDs validated')
		if n == 0:
			print('Query are not in the correct HGVS Format')
		return validated_queries

	@staticmethod
	def validate_coord(queries):
		'''
		standardizes input coordinate and checks formatting
		'''
		# q = 'chr11:5226778C>T'
		coord_fmt = r'(chr[0-9]*:\d*(A|T|C|G)>(A|T|C|G))'
		validated_queries = []
		for q in set(queries):
			if re.search(coord_fmt, q):
				validated_queries.append(re.search(coord_fmt, q).groups()[0])
			else:
				print(q)
		n = len(validated_queries)
		print(f'{n} out of {len(queries)} Coordinates IDs validated')
		if n == 0:
			print('Query are not in the correct Coordinate + allele Format')
		return validated_queries

	def run_FetchGuides(self, outfile_guides, outfile_be_guides):
		global dh, query
		self.fetch_query_info()
		print('Finding Guides.....')
		for ch, data in self.snv_info.items():

			for d in data:
				try:
					query, tid, eid, strand, ref, alt, feature_annotation, extracted_seq, codons, coord = d
				except ValueError:
					print(f"WARNING: The query below has the wrong number of values to unpack. Needs further investigation:\n{d}")
					continue
				dh = DataHandler(query, strand, ref, alt, feature_annotation, extracted_seq, codons, coord)

				if self.be_request != 'off':
					guides, BEguides = dh.get_Guides(self.search_params, self.BE_search_params)
				else:
					guides, BEguides = dh.get_Guides(self.search_params)

				if len(BEguides['gRNA']) > 0:
					if len(self.all_BE.keys()) == 0:
						for k, v in BEguides.items():
							self.all_BE[k] = v
					else:
						for k, v in BEguides.items():
							self.all_BE[k] += v
				if len(guides['gRNA']) > 0:
					if len(self.all_guides.keys()) == 0:
						for k, v in guides.items():
							self.all_guides[k] = v
					else:
						for k, v in guides.items():
							self.all_guides[k] += v

						print(len((guides['gRNA'])), ' guides found for ', query)
				else:
					print(f"No guides found for the query {query}")

		guidedf, BEdf = None, None

		if len(self.all_guides.keys()) != 0:
			guidedf = self.write_guide_csv(self.all_guides, outfile_guides)
			self.clininfo_flag = True
			# self.add_clininfo()

		if len(self.all_BE.keys()) != 0:
			BEdf = self.write_guide_csv(self.all_BE, outfile_be_guides)
			self.clininfo_flag = True

		return {'all_variant': self.all_variant,
		        'all_gene': self.all_gene,
		        'guide_table': guidedf,
		        'BE_table': BEdf}


def main():
	# SNAKEMAKE IMPORTS
	# === Inputs ===
	input_file = str(snakemake.input.query_manifest)
	fasta_path = str(snakemake.input.assembly_path)
	# === Outputs ===
	guides_report = str(snakemake.output.guides_report_out)
	guide_search_params_path = str(snakemake.output.guide_search_params)
	snv_site_info_path = str(snakemake.output.snv_site_info)
	# === Params ===
	resultsfolder = set_export(str(snakemake.params.main_out))
	gene_report = f"{resultsfolder}/{str(snakemake.params.gene_report)}"
	variant_report = f"{resultsfolder}/{str(snakemake.params.variant_report)}"
	be_report = f"{resultsfolder}/{str(snakemake.params.be_report)}"
	#   == Processed tables branch ==
	datadir = str(snakemake.params.support_tables)
	annote_path = str(snakemake.params.annote_path)
	# == Editor Parameters
	editors_path = str(snakemake.params.editors)
	base_editors_path = str(snakemake.params.base_editors)
	#   == Run Parameters ==
	qtype = str(snakemake.params.qtype)
	be_request = str(snakemake.params.be_request)
	editor_request = str(snakemake.params.editor_request)
	# == DEBUG BLOCK ==
	# qtype = 'hgvs'
	# BEmode = 'off'
	# editor = 'all'
	# == == ==

	# == Input Setup ==
	df = pd.read_csv(input_file)
	queries = list(df.iloc[:, 0])

	# == Editors / BEs Setup ==
	with open(editors_path, 'rb') as edfile:
		editors = pickle.load(edfile)
	with open(base_editors_path, 'rb') as befile:
		base_editors = pickle.load(befile)

	# == Report processed input variables ==
	print(f"""
	Currently running fetchGuides.py
	INPUT VARIABLES:
		Queries:\n{queries}
		Query Type: {qtype}
		be_request: {be_request}
		editor_request: {editor_request}
	PATH TO REFERENCE:
		-> {fasta_path}
	SUPPORT DATA DIRECTORY:
		-> {datadir}
	OUTPUTS TO:
		--> {resultsfolder}
	""")

	# == Get query items ==
	fg = Fetch_Guides(queries,
	                  qtype,
	                  editor_request,
	                  be_request,
	                  editors,
	                  base_editors,
	                  datadir,
	                  fasta_path,
	                  annote_path
	                  )
	# == Set up object and run core methods ==
	exports = fg.run_FetchGuides(guides_report, be_report)

	# == Export Intermediate files ==
	fg.write_snv_site_info(snv_site_info_path)
	fg.write_gsearch_params(guide_search_params_path)

	# == Export Variant and Gene tables ==
	fg.add_clininfo(gene_report, variant_report)


if __name__ == "__main__":
	main()
